import pygame
from baopig.io import LOGGER, mouse, keyboard
from baopig.documentation import ApplicationExit, Selector
from baopig.documentation import Focusable as FocusableDoc
from baopig.documentation import HoverableByMouse as HoverableByMouseDoc
from baopig.documentation import LinkableByMouse as LinkableByMouseDoc
from baopig.documentation import Runable as RunableDoc
from baopig.documentation import Validable as ValidableDoc
from baopig.time.timer import RepeatingTimer
from .widget import Widget


class Validable(ValidableDoc):

    def __init__(self, catching_errors=False):

        self._catching_errors = catching_errors

    catching_errors = property(lambda self: self._catching_errors)

    def handle_validate(self):
        """ Stuff to do when validate is called """

    def validate(self):
        if self._catching_errors:
            try:
                self.handle_validate()
            except ApplicationExit as e:
                raise e
            except Exception as e:
                LOGGER.warning(f"Error while executing {self} validation: {e}")
        else:
            self.handle_validate()


class Runable(RunableDoc, Widget):

    def __init__(self, parent, **kwargs):
        Widget.__init__(self, parent, **kwargs)

        self._is_running = False

    is_running = property(lambda self: self._is_running)

    def set_running(self, val):
        self._is_running = bool(val)


class HoverableByMouse(HoverableByMouseDoc, Widget):  # TODO : auomatic update when scroll

    def __init__(self, parent, **kwargs):
        Widget.__init__(self, parent, **kwargs)

        self._is_hovered = False
        self._indicator = None

        # NOTE : these signals are necessary for Indicator
        self.create_signal("HOVER")
        self.create_signal("UNHOVER")

        self.signal.HOVER.connect(self.handle_hover, owner=self)
        self.signal.UNHOVER.connect(self.handle_unhover, owner=self)

        # def drop_hover():
        #     if self._is_hovered:
        #         mouse.update_hovered_widget()
        #
        # self.signal.HIDE.connect(drop_hover, owner=self)
        # self.signal.SLEEP.connect(drop_hover, owner=self)
        #
        # def drop_hover_on_kill():
        #     if self._is_hovered:
        #         mouse.update_hovered_widget()
        #
        # self.signal.KILL.connect(drop_hover_on_kill, owner=self)
        #
        # def check_hover_gain():
        #     if self.collidemouse():
        #         mouse.update_hovered_widget()
        #
        # self.signal.SHOW.connect(check_hover_gain, owner=self)
        # self.signal.WAKE.connect(check_hover_gain, owner=self)
        #
        # def check_hover():
        #     if self._is_hovered and not self.collidemouse():
        #         mouse.update_hovered_widget()
        #     elif self.collidemouse():
        #         mouse.update_hovered_widget()
        #
        # self.signal.MOTION.connect(check_hover, owner=self)
        # self.signal.RESIZE.connect(check_hover, owner=self)
        #
        # if self.collidemouse():
        #     mouse.update_hovered_widget()

    indicator = property(lambda self: self._indicator)
    is_hovered = property(lambda self: self._is_hovered)


# ...


class LinkableByMouse(LinkableByMouseDoc, HoverableByMouse):

    def __init__(self, parent, **kwargs):
        HoverableByMouse.__init__(self, parent, **kwargs)

        self.is_linked = False  # non-protected field, manipulated by the mouse

    def handle_link(self):
        """Stuff to do when the widget gets linked"""

    def handle_link_motion(self, rel):
        """Stuff to do when the widget'link has changed"""

    def handle_unlink(self):
        """Stuff to do when the widget's link is over"""

    def unlink(self):
        """Send a request for unlinking this widget"""
        if self.is_linked:
            mouse._unlink()


class Focusable(FocusableDoc, LinkableByMouse):

    def __init__(self, parent, **kwargs):

        LinkableByMouse.__init__(self, parent, **kwargs)

        self._is_focused = False

        self.signal.HIDE.connect(self.defocus, owner=None)
        self.signal.SLEEP.connect(self.defocus, owner=None)

    is_focused = property(lambda self: self._is_focused)

    def defocus(self):
        """ Send a request for defocusing this widget """
        if self.is_focused:
            self.scene.focus(None)

    def handle_defocus(self):
        """ Called when the widget looses the focus """

    def handle_focus(self):
        """ Called when the widget receives the focus """

    # KEY EVENTS LISTENING

    is_listening_keyevents = property(lambda self: self._is_focused)

    def handle_keydown(self, key):
        """ Called when a key is pressed """

        if keyboard.mod.ctrl:  # TODO : ctrl or cmd
            if key in (pygame.K_a, pygame.K_c, pygame.K_v, pygame.K_x):
                if not isinstance(self, Selector):  # if not isinstance(self, Selector)
                    selector = self.parent
                    while not isinstance(selector, Selector):  # not infinite since Scene is a Selector
                        selector = selector.parent
                    selector.handle_keydown(key)  # parent is now the closest Selector parent

        if key == pygame.K_TAB:
            self.handle_tab()

    def handle_tab(self):
        """
        Give the focus to the next Focusable (ranked by position) in parent
        If maj is pressed, gives the focus to the previous Focusable
        """
        all_focs = []
        for child in self.parent.children:
            if isinstance(child, Focusable):
                if child.is_visible:
                    all_focs.append(child)

        # TAB       -> focus the next Focusable inside this widget's parent
        # Maj + TAB -> focus the previous Focusable inside this widget's parent
        d = 1 if keyboard.mod.maj == 0 else -1

        if len(all_focs) > 1:
            all_focs.sort(key=lambda c: (c.rect.top, c.rect.left))
            self.scene.focus(all_focs[(all_focs.index(self) + d) % len(all_focs)])


class DraggableByMouse(LinkableByMouse):
    """
    Class for widgets who want to be moved by mouse
    """

    def handle_link_motion(self, rel):
        self.move(*rel)


class RepetivelyAnimated(Widget):  # TODO : rework default anitmations
    """
    A RepetivelyAnimated is a widget who blinks every interval seconds

    Exemple :

        class Lighthouse(RepetivelyAnimated):

    """

    def __init__(self, parent, interval, **kwargs):
        """
        The widget will appear and disappear every interval seconds
        :param interval: the time between appear and disappear
        """

        Widget.__init__(self, parent, **kwargs)

        assert isinstance(interval, (int, float)), "interval must be a float or an integer"

        self.interval = interval

        def blink():
            if self.is_visible:
                self.hide()
            else:
                self.show()

        self._countdown_before_blink = RepeatingTimer(interval, blink)
        self._need_start_animation = False
        self.signal.SLEEP.connect(self.handle_sleep, owner=None)
        self.signal.WAKE.connect(self.handle_wake, owner=None)
        self.signal.KILL.connect(self._countdown_before_blink.cancel, owner=None)

    def handle_sleep(self):

        self._need_start_animation = self._countdown_before_blink.is_running
        self._countdown_before_blink.cancel()

    def start_animation(self):

        if self.is_asleep:
            self._need_start_animation = True
            return

        if self._countdown_before_blink.is_running:
            self._countdown_before_blink.cancel()
        self._countdown_before_blink.start()

    def stop_animation(self):

        if self.is_asleep:
            self._need_start_animation = False
            return

        self._countdown_before_blink.cancel()

    def handle_wake(self):

        if self._need_start_animation:
            self._countdown_before_blink.start()
        self._need_start_animation = False



from baopig import *


class UT_Text_Zone(Zone):

    def __init__(self, *args, **kwargs):
        Zone.__init__(self, *args, **kwargs)

        Layer(self, name="zones_layer", spacing=10)
        z1 = Zone(self, size=("100%", 80), background_color=(150, 150, 150), spacing=5, padding=5)
        z2 = Zone(self, size=("100%", 98), background_color=(150, 150, 150), spacing=5, padding=5)
        sv = ScrollView(self, size=("100%", 300), background_color=(150, 150, 150), padding=5)
        z3 = Zone(sv, spacing=10)
        z4 = Zone(self, size=("100%", 120), background_color=(150, 150, 150), spacing=5, padding=5)
        z5 = Zone(self, size=("100%", 120), background_color=(150, 150, 150), spacing=5, padding=5)
        self.default_layer.pack()

        # Z1
        hello = Text(parent=z1, text="Hello world\nIamaverylongword,canyoureadmecorrectly?"
                                     "\nWhat do you want to do ?", max_width=250, padding=5)
        z1.set_style_for(Text, font_file="algerian", font_italic=True)
        Text(parent=z1, text="Hello world\nIamaverylongword,canyoureadmecorrectly?"
                             "\nWhat do you want to do ?", max_width=hello.rect.width)
        z1.pack(axis="horizontal")

        # Z2
        z2.set_style_for(Text, font_file="monospace")
        z2.set_style_for(Text, font_height=60)
        z2.set_style_for(Text, font_color=(0, 150, 0))
        text = Text(z2, max_width=300, font_height=11, font_file="arialrounded",
                    text="- Bonjour à tous.\n"
                         "- Bonjour monsieur. Comment allez-vous ?\n"
                         "- Très bien, merci. Nous allons commencer. \"L'hypoténuse et l'hippocampe "
                         "se préparaient à pique-niquer, lorsque...\"\n"
                         "- Monsieur, j'avais oublié, j'ai un rendez-vous !\n"
                         "- Eh bien, filez, ça ira pour cette fois.\n"
                         "- Merci monsieur !\n"
                         "Et il partit. (vert fonce)")
        text.set_background_color((255, 255, 255, 128))
        text.font.config(color=(10, 50, 30))
        TextEdit(z2, text="Green", size=text.rect.size, font_file="passeroone")
        z2.pack(axis="horizontal")

        # Z3
        def load_sys_fonts():
            calls_index = 1
            row = 0
            col = 0
            fonts_zone.set_pos(top=-1)  # placed before the button
            fonts_zone.set_style_for(Text, font_height=18)
            fonts_zone.set_style_for(Text, font_color=(0, 0, 0))
            for file in pygame.font.get_fonts():
                # if file.endswith("gras") or file.endswith("italic") or file.endswith("italique") \
                #         or file.endswith("oblique"):
                #     continue
                Text(fonts_zone, text=file[:-4] if file.endswith(".ttf") else file, font_file=file,
                     row=row, col=col)
                col += 1
                if col == 4:
                    col = 0
                    row += 1
                if row > 100 * calls_index:
                    fonts_zone.pack(adapt=True)
                    z3.pack(adapt=True)
                    calls_index += 1
                    yield "STILL FONTS LEFT"
            fonts_zone.pack(adapt=True)
            z3.pack(adapt=True)
            yield "DONE"

        load_sys_fonts_generator = load_sys_fonts()

        def preview_fonts():
            try:
                print(next(load_sys_fonts_generator))
            except StopIteration:
                pass

        self.b = Button(z3, "Click here to see more fonts provided by your OS", command=preview_fonts,
                        width=sv.content_rect.width)
        fonts_zone = Zone(z3, height=0)
        GridLayer(fonts_zone, nbcols=4, col_width=int(sv.rect.width / 4), spacing=(5, 2))
        z3.pack(adapt=True)

        # Z4
        Text(z4, "max_width:150", max_width=150)
        Text(z4, "max_width:75", max_width=75)
        Text(z4, "padding:5", padding=5)
        Text(z4, "1\n2\n3", max_width=85, padding=10, spacing=10)
        z4.pack()

        # Z5
        z5.set_style_for(Text, font_height=15)
        widget = Text(z5, "Hello world", background_color="green4")
        widget.set_text("Hello world and everyone else")  # -> here, the font_height will be reduced
        Text(z5, "Hello world", background_color="green4")
        Text(z5, "Hello world", background_color="green4", max_width=100)
        Text(z5, "Hello world and everyone else", background_color="green4", max_width=100)
        z5.pack()


# For the PresentationScene import
ut_zone_class = UT_Text_Zone

if __name__ == "__main__":
    from baopig.prefabs.testerscene import TesterScene

    app = Application()
    TesterScene(app, ut_zone_class)
    app.launch()

"""
# Default code for Try it yourself
from baopig import *


class UT_Text_Zone(Zone):

    def __init__(self, *args, **kwargs):
        Zone.__init__(self, *args, **kwargs)

        Text(parent=self, padding=5, max_width=self.rect.width,
             font_file="passeroone", font_height=50,
             text="Hello world\n"
                  "I am a very long text\n"
                  "but i'm sure que tu peux me lire sans probleme")


app = Application(size=(700, 700))
main_zone = UT_Text_Zone(Scene(app), size=("90%", "90%"), sticky="center")
app.launch()
"""





import sys
import os
import time
import threading
import pygame
from baopig.pybao.objectutilities import History, Object
from baopig.time.timer import RepeatingTimer
from baopig.lib import paint_lock
from .thread import ExtraThread, LOGGER


class PainterThread(ExtraThread):

    def __init__(self, app):

        ExtraThread.__init__(self, app)

        basename = os.path.splitext(os.path.basename(sys.argv[0]))[0]
        self.out_directory = os.path.abspath("out") + os.path.sep + basename + os.path.sep
        self._is_recording = False
        self.record_index = 1

        self._required_fps = None

        # if self.app._debug_averagefps:
        def _tick_fps():
            # being a deque, it manages its data itself
            self.fps_history.append(self.screenupdates_during_current_second)
            self.screenupdates_during_current_second = 0

        self.screenupdates_during_current_second = 0
        self.fps_history = History(seq=[], maxlen=500)
        self.fps_history_updater = RepeatingTimer(1, _tick_fps)

        # The clock is the object who freeze enough the thread in order to have maximum app.required_fps FPS
        self.clock = pygame.time.Clock()
        self._can_draw = threading.Event()

    def __del__(self):

        # if self.app._debug_averagefps:
        self.fps_history_updater.cancel()

    is_recording = property(lambda self: self._is_recording)
    required_fps = property(lambda self: self._required_fps)

    def get_current_fps(self):
        """

        :return:
        """
        # if self.app._debug_averagefps:
        if len(self.fps_history) > 0:
            return self.fps_history[-1]
        else:
            return None
        # else:
        #     return self.clock.get_fps()  # TODO : fps track even with infinite fps

    def init(self):

        # if self.app._debug_averagefps:
        self.fps_history_updater.start()

    def screenshot(self):

        os.makedirs(self.out_directory, exist_ok=True)
        screenshot = self.app.display.copy()
        name = time.strftime("%Y.%m.%d-%Hh%M-%S.png", time.localtime())
        print(self.out_directory + name)
        pygame.image.save(screenshot, self.out_directory + name)

    def set_fps(self, fps):

        assert isinstance(fps, int) and fps > 0 or fps is None
        self._required_fps = fps

    def stop(self):

        with paint_lock:
            super().stop()
            # if self.app._debug_averagefps:
            self.fps_history_updater.cancel()

    def start_recording(self, only_at_change=False):

        os.makedirs(self.out_directory, exist_ok=True)
        self._is_recording = Object(only_at_change=only_at_change)
        LOGGER.info("Start recording" + (" (only at display updates)" if only_at_change else ""))

    def stop_recording(self):

        self._is_recording = False
        LOGGER.info("Stop recording")

    def update(self):

        self._can_draw.wait()

        try:

            # Drawings
            with paint_lock:
                try:
                    self.app.focused_scene._container_paint()
                except Exception as e:
                    LOGGER.exception(e)

            # FPS Tracer
            """if self.fps_label.is_visible:
                for i, required_fps in enumerate(self.fps_history):
                    color = 127
                    pygame.draw.line(self.display,
                                     (color, color, color),
                                     (i + 10, self.rect.bottom - 10),
                                     (i + 10, self.rect.bottom - 10 - required_fps * 10))"""

            # launch time
            if self.app._debug_launchtime and self.app.launch_time is not None:
                import time
                LOGGER.info("{} launched in {} seconds".format(self.app.name, time.time() - self.app.launch_time))
                self.app.launch_time = None

            # record
            if self.is_recording:
                if not self.is_recording.only_at_change:
                    pygame.image.save(self.app.display, self.out_directory + f"record_{self.record_index:0>3}.png")
                    self.record_index += 1

            # FPS
            # if self.app._debug_averagefps:
            self.screenupdates_during_current_second += 1
            # NOTE : Pour mieux tester les FPS, on ne fait pas ticker l'horloge
            if self.required_fps is not None:
                self.clock.tick(self.required_fps)  # keep the game running slower than the given FPS
            self._can_draw.clear()

        except pygame.error as e:
            if e.__str__() == "video system not initialized":
                self.stop()
                LOGGER.error("The window have been closed")
            else:
                raise e



from baopig import *


class HoverableRectangle(Rectangle, HoverableByMouse):
    STYLE = Rectangle.STYLE.substyle()
    STYLE.modify(
        color="blue"
    )

    def __init__(self, parent, **kwargs):
        Rectangle.__init__(self, parent, **kwargs)

        self.is_main = True  # needed in HoverableByMouse.__init__()
        HoverableByMouse.__init__(self, parent)

    def handle_hover(self):
        self.set_color("yellow" if self.is_main else "yellow4")

    def handle_unhover(self):
        self.set_color("blue" if self.is_main else "blue4")

    def set_main(self, val):
        self.is_main = val
        if self.is_hovered:
            self.handle_hover()
        else:
            self.handle_unhover()


class TestingZone(Zone, Focusable):

    def __init__(self, parent, **kwargs):
        Zone.__init__(self, parent, **kwargs)
        Focusable.__init__(self, parent)
        self._controlled_ref = lambda: None
        self.highlighter = Highlighter(self, target=self, visible=False, border_width=4)

    controlled = property(lambda self: self._controlled_ref())

    def handle_defocus(self):
        self.highlighter.hide()

    def handle_focus(self):
        self.highlighter.show()

    def handle_keydown(self, key):

        if key == pygame.K_a:
            if self.controlled is not None:
                self.controlled.set_main(False)
            self._controlled_ref = HoverableRectangle(self, sticky="center").get_weakref()

        elif self.controlled is None:
            return
        vel = 5
        if self.controlled.is_awake:
            if key == pygame.K_LEFT:
                self.controlled.move(dx=-vel)
            elif key == pygame.K_RIGHT:
                self.controlled.move(dx=vel)
            elif key == pygame.K_UP:
                self.controlled.move(dy=-vel)
            elif key == pygame.K_DOWN:
                self.controlled.move(dy=vel)

        if key == pygame.K_s:
            self.controlled.show()
        elif key == pygame.K_h:
            self.controlled.hide()

        elif key == pygame.K_w:
            self.controlled.wake()
        elif key == pygame.K_l:
            self.controlled.sleep()

        elif key == pygame.K_k:
            self.controlled.kill()
            last_child = self.default_layer[-1]
            if last_child != self.highlighter:
                self._controlled_ref = last_child.get_weakref()
                self.controlled.set_main(True)

        elif key == pygame.K_r:
            if self.controlled.w == 30:
                self.controlled.resize(40, 40)
            elif self.controlled.w == 40:
                self.controlled.resize(20, 20)
            else:
                self.controlled.resize(30, 30)

        elif key == pygame.K_n:
            self.controlled.set_touchable_by_mouse(False)  # TODO : tests


class UT_Hoverable_Zone(Zone):
    def __init__(self, *args, **kwargs):
        Zone.__init__(self, *args, **kwargs)

        Layer(self, name="zones_layer", spacing=10)
        z1 = Zone(self, size=(90, 50), background_color=(150, 150, 150), padding=10, spacing=10)
        z2 = Zone(self, size=("100%", 250))
        self.default_layer.pack()

        # Z1
        hr = HoverableRectangle(z1)
        light = Rectangle(z1, color="black")
        hr.signal.HOVER.connect(lambda: light.set_color("red3"), owner=light)
        hr.signal.UNHOVER.connect(lambda: light.set_color("black"), owner=light)
        z1.default_layer.pack(axis="horizontal")

        # Z2
        TestingZone(z2, size=("50%", "100%"), background_color=(150, 150, 150))
        text_zone = Zone(z2, size=("50%", "100%"))
        z2.default_layer.pack(axis="horizontal")
        Text(text_zone, max_width=text_zone.rect.w,
             text="Focus the zone at left in order to use it with your keyboard.\n"
                  "Use the arrows to control the light blue rect\n"
                  "Press the following keys to execute actions over the light blue rect:\n"
                  " - a : Create a new widget\n"
                  " - s : Show the widget\n"
                  " - h : Hide the widget\n"
                  " - w : Wake up the widget\n"
                  " - l : Asleep the widget\n"
                  " - k : Kill the widget\n"
                  " - n : Set the widget non-touchable\n"
             )

    def load_sections(self):
        self.parent.add_section(
            title="mouse manages mouse.hovered_widget",
            tests=[
                "When the mouse moves, the hovered_widget is updated",
            ]
        )
        self.parent.add_section(
            title="HoverableByMouse can manage mouse.hovered_widget",
            tests=[
                "When a hovered HoverableByMouse disappears (hide, sleep, kill), it drops the hover",
                "When a hovered HoverableByMouse moves or get resized, if needed, it drops the hover",
                "When a HoverableByMouse appears (creation, show, wake), if needed, it gains the hover",
                "When a HoverableByMouse moves or get resized, if needed, it gains the hover",
            ]
        )
        self.parent.add_section(
            title="signal.HOVER & signal.UNHOVER",
            tests=[
                "When a HoverableByMouse is just hovered, the signal HOVER is emitted",
                "When a hovered HoverableByMouse is unhovered, the signal UNHOVER is emitted",
            ]
        )


# For the PresentationScene import
ut_zone_class = UT_Hoverable_Zone

if __name__ == "__main__":
    from baopig.prefabs.testerscene import TesterScene
    app = Application()
    TesterScene(app, ut_zone_class)
    app.launch()




from baopig.pybao.objectutilities import WeakTypedList
from .utilities import MarginType
from .widget import Widget, Communicative


class Layer(Communicative):
    """
    A Layer is a manager who contains some of its container's children.
    Every widget is stored in one of its parent's layers.
    The positions of widgets inside a layer define the overlay : first behind, last in front.
    Each layer can be overlaid in the foreground, the main ground or the background.
    2 layers from the same ground are overlaid depending on their weight : a weight of 0 means it needs
    to stand behind a layer with weight 6. The default weight is 2.
    """

    def __init__(self, container, *filter_cls, name=None, level=None, weight=None, padding=None, spacing=None,
                 default_sortkey=None, sort_by_pos=False, touchable=True, maxlen=None, adaptable=False):
        """
        :param container: the Container who owns the layer
        :param name: an unic identifier for the layer
        :param filter_cls: a class or list of class from wich every layer's widget must herit
        :param level: inside the container's layers : lowest behind greatest in front, default to MAINGROUND
        :param weight: inside the layer's level : lowest behind greatest in front, default to 2
        :param padding: space between the widgets and the container. If None, set to the container's padding
        :param spacing: space between 2 widgets. If None, set to the container's spacing
        :param default_sortkey: default key fo layer.sort(). if set, at each append, the layer will be sorted
        :param sort_by_pos: if set, the default sortkey will be a function who sort children by y then x
        :param touchable: children of non-touchable layer are not hoverable
        :param maxlen: the maximum numbers of children the layer can contain
        """  # TODO : start_pos

        if name is None:
            name = "UnnamedLayer{}".format(len(container.layers))
        if not filter_cls:
            filter_cls = [Widget]
        if level is None:
            level = container.layers_manager.DEFAULT_LEVEL
        if weight is None:
            weight = 2
        if padding is None:
            padding = container.padding  # Same object
        else:
            padding = MarginType(padding)
        if spacing is None:
            spacing = container.spacing  # Same object
        else:
            spacing = MarginType(spacing)
        if sort_by_pos:
            assert default_sortkey is None
            default_sortkey = lambda c: (c.rect.top, c.rect.left)

        for filter_class in filter_cls:
            assert issubclass(filter_class, Widget), filter_class
        assert isinstance(name, str), name
        assert name not in container.layers, name
        assert level in container.layers_manager.levels, level
        assert isinstance(weight, (int, float)), weight
        assert isinstance(padding, MarginType), padding
        assert isinstance(spacing, MarginType), spacing
        if default_sortkey is not None:
            assert callable(default_sortkey), default_sortkey
        if maxlen is not None:
            assert isinstance(maxlen, int), maxlen

        Communicative.__init__(self)

        # NOTE : adaptable, container, name, touchable and level are not editable, because you
        #        need to know what kind of layer you want since its creation
        self._is_adaptable = bool(adaptable)
        self._widgets = WeakTypedList(*filter_cls)
        self._container = container
        self._filter = filter_cls
        self._name = name
        self._level = level
        self._weight = weight
        self._padding = padding
        self._spacing = spacing
        self.default_sortkey = default_sortkey  # don't need protection
        self._layer_index = None  # defined by container.layers
        self._layers_manager = container.layers_manager
        self._maxlen = maxlen
        self._touchable = bool(touchable)

        self.layers_manager._add_layer(self)

    def __add__(self, other):
        return self._widgets + other

    def __bool__(self):
        return bool(self._widgets)

    def __contains__(self, item):
        return self._widgets.__contains__(item)

    def __getitem__(self, item):
        return self._widgets.__getitem__(item)

    def __iter__(self):
        return self._widgets.__iter__()

    def __len__(self):
        return self._widgets.__len__()

    def __repr__(self):
        return "{}(name:{}, index:{}, filter_cls:{}, touchable:{}, level:{}, weight:{}, children:{})".format(
            # "Widgets" if self.touchable else "",
            self.__class__.__name__, self.name, self._layer_index, self._filter, self.touchable,
            self.level, self.weight, self._widgets)

    spacing = property(lambda self: self._spacing)
    container = property(lambda self: self._container)
    is_adaptable = property(lambda self: self._is_adaptable)
    layer_index = property(lambda self: self._layer_index)
    layers_manager = property(lambda self: self._layers_manager)
    level = property(lambda self: self._level)
    maxlen = property(lambda self: self._maxlen)
    name = property(lambda self: self._name)
    padding = property(lambda self: self._padding)
    touchable = property(lambda self: self._touchable)
    weight = property(lambda self: self._weight)

    def accept(self, widget):

        if self.maxlen and self.maxlen <= len(self._widgets):
            return False
        return self._widgets.accept(widget)

    def add(self, widget):
        """
        WARNING : This method should only be called by the LayersManager: cont.layers_manager
        You can override this function in order to define special behaviors
        """
        if self.maxlen and self.maxlen <= len(self._widgets):
            raise PermissionError("The layer is full (maxlen:{})".format(self.maxlen))

        self._widgets.append(widget)
        if self.default_sortkey:
            self.sort()

        if self.is_adaptable:
            self.container.adapt(self)

    def clear(self):

        for widget in tuple(self._widgets):
            widget.kill()

    def get_visible_widgets(self):
        for widget in self._widgets:
            if widget.is_visible:
                yield widget

    visible = property(get_visible_widgets)

    def index(self, widget):
        return self._widgets.index(widget)

    def kill(self):

        self.clear()
        self.layers_manager._remove_layer(self)

    def move_widget1_behind_widget2(self, widget1, widget2):
        assert widget1 in self._widgets, "{} not in {}".format(widget1, self)
        assert widget2 in self._widgets, "{} not in {}".format(widget2, self)
        self.overlay(self.index(widget2), widget1)

    def move_widget1_in_front_of_widget2(self, widget1, widget2):
        assert widget1 in self, "{} not in {}".format(widget1, self)
        assert widget2 in self, "{} not in {}".format(widget2, self)
        self.overlay(self.index(widget2) + 1, widget1)
        # self._remove(widget1)
        # super().insert(self.index(widget2) + 1, widget1)
        # self._warn_change(widget1.hitbox)

    def overlay(self, index, widget):
        """
        Move a widget at index
        """

        assert widget in self._widgets
        self._widgets.remove(widget)
        self._widgets.insert(index, widget)
        self.container._warn_change(widget.hitbox)

    def pack(self, key=None, axis="vertical", spacing=None, padding=None, start_pos=(0, 0)):
        """
        Place children on one row or one column, sorted by key (default : pos)
        axis can either be horizontal or vertical
        NOTE : if motivated, can add 'sticky' param, which places the packed children from a corner or the center
        """
        if key is None:
            key = lambda o: (o.rect.top, o.rect.left)
        if spacing is None:
            spacing = self._spacing
        if padding is None:
            padding = self._padding
        if not isinstance(spacing, MarginType):
            spacing = MarginType(spacing)
        if not isinstance(padding, MarginType):
            padding = MarginType(padding)

        sorted_children = sorted(self, key=key)

        left, top = padding.left + start_pos[0], padding.top + start_pos[1]
        for widget in sorted_children:
            if widget.has_locked("pos"):
                raise PermissionError("Cannot pack a layer who contains locked children")
            widget.set_pos(topleft=(left, top))

            if axis == "horizontal":
                left = widget.hitbox.right + spacing.left
            elif axis == "vertical":
                top = widget.hitbox.bottom + spacing.top
            else:
                raise ValueError(f"axis must be either 'horizontal' or 'vertical', not {axis}")

    def remove(self, widget):
        """
        WARNING : This method should only be called by the LayersManager: cont.layers_manager
        You can override this function in order to define special behaviors
        """
        self._widgets.remove(widget)

        if self.is_adaptable:
            self.container.adapt(self)

    def set_filter(self, filter_cls):

        self._widgets.set_ItemsClass(filter_cls)

    def set_maxlen(self, maxlen):

        assert isinstance(maxlen, int) and len(self._widgets) <= maxlen
        self._maxlen = maxlen

    def set_weight(self, weight):

        assert isinstance(weight, (int, float))
        self._weight = weight
        self.layers_manager.sort_layers()

    def sort(self, key=None):
        """
        Permet de trier les enfants d'un layer selon une key
        Cette fonction ne deplace pas les enfants, elle ne fait que changer leur
        superpositionnement
        """
        if key is None:
            key = self.default_sortkey
        if key is None:  # No sort key defined
            return
        self._widgets.sort(key=key)



from baopig import mouse
from baopig import Zone, Handler_SceneClose, Text, DynamicText, Highlighter


# --- DEBUG ---
class DebugZone(Zone, Handler_SceneClose):

    def __init__(self, scene):

        Zone.__init__(self, scene, size=scene.size, background_color=(0, 0, 0, 0), layer=scene.debug_layer,
                      touchable=False)

        self.set_style_for(Text, font_color="black", font_file=None)

        self._pointed = None
        self.highlighter = None

        h = 200
        self.debug_zone = Zone(
            parent=self,
            size=(self.rect.w - 10, h),
            background_color=(255, 255, 255, 145),
            name=self.name + " -> debug_zone",
            bottomleft=(5, -5), refloc="bottomleft"
        )

        presentators_zone = Zone(
            parent=self.debug_zone,
            pos=(5, 5),
            size=(120, self.debug_zone.rect.h),
            name=self.debug_zone.name + " -> presentators_zone"
        )
        trackers_zone = Zone(
            parent=self.debug_zone,
            pos=(presentators_zone.rect.right, 5),
            size=(4000, self.debug_zone.rect.h),
            name=self.debug_zone.name + " -> trackers_zone"
        )

        if True:
            # FPS TRACKER
            Text(presentators_zone, text="FPS : ")
            DynamicText(trackers_zone, get_text=lambda: self.parent.painter.get_current_fps())

            # MOUSE TRACKER
            Text(presentators_zone, text="Mouse pos : ")
            DynamicText(trackers_zone, get_text=lambda: mouse.pos)

            Text(presentators_zone, text="Pointed widget : ")
            Text(trackers_zone)

            # CLASS TRACKER
            Text(presentators_zone, text="- class : ")
            DynamicText(trackers_zone, get_text=lambda: self._pointed.__class__.__name__ if self._pointed else None)

            # NAME TRACKER
            Text(presentators_zone, text="- name : ")
            DynamicText(trackers_zone, get_text=lambda: self._pointed.name if self._pointed else None)

            # HITBOX TRACKER
            Text(presentators_zone, text="- hitbox : ")
            DynamicText(trackers_zone, get_text=lambda: self._pointed.hitbox if self._pointed else None)

            # HITBOX TRACKER
            Text(presentators_zone, text="- abs_hitbox : ")
            DynamicText(trackers_zone, get_text=lambda: self._pointed.abs_hitbox if self._pointed else None)

            # PARENT TRACKER
            Text(presentators_zone, text="- parent : ")
            DynamicText(trackers_zone, get_text=lambda: self._pointed.parent if self._pointed else None)

        presentators_zone.pack()
        trackers_zone.pack()

        self.print_text = Text(
            parent=self.debug_zone,
            text="This text is aimed to debug",
            pos=(5, self.debug_zone.rect.h - 5 - 15),
            name=self.debug_zone.name + " -> print_text"
        )

        self.parent.signal.RESIZE.connect(self.handle_scene_resize, owner=self)
        mouse.signal.MOUSEMOTION.connect(self.update_pointed_outline, owner=self)

        self.update_pointed_outline()
        self.debug_zone.adapt(self.debug_zone.default_layer, horizontally=False)

    is_debugging = property(lambda self: self.is_awake)

    def handle_scene_close(self):

        self.kill()  # TODO : not kill

    def handle_scene_resize(self):

        self.resize(*self.scene.size)
        self.debug_zone.resize_width(self.rect.w - 10)
        self.update_pointed_outline()

    def print(self, obj):

        self.print_text.set_text(str(obj))
        self.debug_zone.adapt(self.debug_zone.default_layer, horizontally=False)

    def start_debugging(self):

        if self.is_debugging:
            return

        self.wake()

    def stop_debugging(self):

        if not self.is_debugging:
            return

        self.sleep()

    def toggle_debugging(self):

        if self.is_debugging:
            self.stop_debugging()
        else:
            self.start_debugging()

    def update_pointed_outline(self):

        def collidemouse(widget):
            return widget.is_visible and widget.abs_hitbox.collidepoint(mouse.pos)

        def get_pointed_widget(cont):

            for layer in reversed(tuple(cont.layers_manager.touchable_layers)):
                assert layer.touchable
                for child in reversed(layer):
                    if child is self:
                        continue
                    if collidemouse(child):
                        if hasattr(child, "children"):
                            return get_pointed_widget(child)
                        return child
            return cont

        if self.is_asleep:
            return

        pointed = get_pointed_widget(self.scene)
        if pointed:
            if self._pointed == pointed:
                return
            self._pointed = pointed
            if self.highlighter is not None:
                self.highlighter.kill()
            self.highlighter = Highlighter(parent=self, target=pointed, name=self.name + ".highlighter")
            self.highlighter.move_behind(self.debug_zone)
        else:
            self._pointed = None

    def wake(self):

        super().wake()
        self.update_pointed_outline()




# NOTE : filename is imagewidget.py because image.py would overlap pygame.image

import pygame
from .widget import Widget


class Image(Widget):

    # TODO : self.tiled instead of parameter in resize()

    def __init__(self, parent, image, w=None, h=None, tiled=False, **kwargs):  # TODO : rework w & h params
        """
        Cree une image

        If w or h parameters are filled, width or height of image argument
        are respectively resized
        """

        assert isinstance(image, pygame.Surface), "image must be a Surface"

        image_size = image.get_size()
        if w is None:
            w = image_size[0]
        if h is None:
            h = image_size[1]
        if image_size != (w, h):
            surface = pygame.transform.scale(image, (w, h))
        else:
            surface = image.copy()

        Widget.__init__(self, parent=parent, surface=surface, **kwargs)
        if self._asked_size[0] is None:
            self._asked_size = (self.rect.w, self._asked_size[1])
        if self._asked_size[1] is None:
            self._asked_size = (self._asked_size[0], self.rect.h)

        self._original = image
        self._tiled = tiled

    def _update_surface_from_resize(self, asked_size):

        if self._tiled:
            super()._update_surface_from_resize(asked_size)

        else:
            self.set_surface(pygame.transform.scale(self._original, asked_size))

    def collidemouse_alpha(self):  # TODO
        raise NotImplemented

    def paint(self):  # TODO : tests

        if self._tiled:

            width, height = self.rect.size
            self.surface.blit(self._original, (0, 0))
            original_w, original_h = self._original.get_size()

            if width > original_w:
                for i in range(int(width / original_w)):
                    self.surface.blit(self.surface, (original_w * (i + 1), 0))

            if height > original_h:
                row = self.surface.subsurface((0, 0, width, original_h)).copy()
                for i in range(int(height / original_h)):
                    self.surface.blit(row, (0, original_h * (i + 1)))




from .layer import Widget, Layer


class LayersManager:

    levels = {
        0: "BACKGROUND",
        1: "MAINGROUND",
        2: "FOREGROUND",
    }
    BACKGROUND = 0
    MAINGROUND = 1
    FOREGROUND = 2
    DEFAULT_LEVEL = MAINGROUND

    def __init__(self, container):

        self._container = container
        self._layers = []
        self._default_layer = None
        self._touchable_layers = []

    container = property(lambda self: self._container)
    default_layer = property(lambda self: self._default_layer)
    layers = property(lambda self: self._layers)
    touchable_layers = property(lambda self: self._touchable_layers)

    background = property(lambda self: self.get_level(self.BACKGROUND))
    foreground = property(lambda self: self.get_level(self.FOREGROUND))
    mainground = property(lambda self: self.get_level(self.MAINGROUND))

    def _add_layer(self, layer):
        """
        This method should only be called by Layer.__init__
        """

        assert len(layer) == 0, "A layer should be empty at its creation"
        assert layer.name not in self._layers
        assert layer.level in self.levels

        if layer.touchable and layer.level == self.MAINGROUND:
            if self._default_layer is None:
                self._default_layer = layer

        self._layers.append(layer)
        self._sort_layers()
        self._touchable_layers = list(layer for layer in self._layers if layer.touchable)  # this preserv the overlay

    def _remove_layer(self, layer):
        """
        This method should only be called by Layer.kill
        """

        assert len(layer) == 0
        self._layers.remove(layer)
        if layer.touchable:
            self._touchable_layers.remove(layer)
        if layer == self.default_layer:
            self._default_layer = None

    def _sort_layers(self):

        layer_index = 0
        for level in sorted(self.levels.keys()):
            for layer in sorted(self.get_ilevel(level), key=lambda layer: layer.weight):
                layer._layer_index = layer_index
                layer_index += 1
        self._layers.sort(key=lambda layer: layer.layer_index)

    @staticmethod
    def accept(child):
        return isinstance(child, Widget)

    def add(self, child):
        """
        This method should only be called by Container.children._add
        You can override this function in order to define special behaviors
        """

        if child.layer is None:
            raise AssertionError

        assert child.layer in self._layers, (child.layer, self.layers)

        child.layer.add(child)

    def get_ilevel(self, level):

        for layer in self._layers:
            if layer.level == level:
                yield layer

    def get_layer(self, layer_name):

        for layer in self._layers:
            if layer.name == layer_name:
                return layer  # if no layer is found, return None
        raise IndexError(f"No layer with name : {layer_name}")

    def get_level(self, level):

        return tuple(self.get_ilevel(level))

    def remove(self, child):

        assert child.layer in self._layers, child.layer
        child.layer.remove(child)

    def set_default_layer(self, layer):

        if isinstance(layer, str):
            layer = self.get_layer(layer)
        self._default_layer = layer

    def find_layer_for(self, child, layer_level=None):

        if layer_level is None:
            layer_level = self.DEFAULT_LEVEL  # TODO : default level is always MAINGROUND

        for layer in self.layers:
            if layer.level == layer_level and layer.accept(child):
                return layer
        return Layer(self.container, level=layer_level, name="automaticlyCreatedLayer")

    def swap_layer(self, child, layer):

        assert layer in self._layers, layer
        if child.is_awake:
            assert child in child.layer, child.layer
            child.layer.remove(child)
            layer.add(child)
        child._layer = layer
        child.send_display_request()





import inspect
from baopig.pybao.objectutilities import Object
from baopig.io.logging import LOGGER
from .documentation import ApplicationExit
from .documentation import Communicative as CommunicativeDoc


class Signal:

    def __init__(self, emitter, id):

        assert isinstance(emitter, Communicative), emitter
        assert isinstance(id, str) and (id == id.upper()), id

        self._emitter = emitter
        self._id = id
        self._connections = []

    def __str__(self):

        return f"Signal(id={self._id}, emitter={self._emitter})"

    def connect(self, command, owner):
        """
        Connect the command to the signal
        When self will emit 'signal', the owner's method 'command' will be executed
        :param command: a method of owner
        :param owner: a Communicative object
        NOTE : The "owner" parameter is very important when it comes to deletion
               When the owner is deleted, this connection is automatically killed
        """
        if not callable(command):
            raise TypeError(f"'{command}' object is not callable")
        for con in self._connections:
            if con.slot is command:
                raise PermissionError(f"This command is already connected to the signal {self._id}")

        conn = Connection(owner, self, command)
        self._connections.append(conn)
        if owner is not None:
            owner._connections.add(conn)

    def disconnect(self, command):

        for con in tuple(self._connections):
            if con.slot == command:
                con.kill()
                return
        raise ValueError(f"This command is not connected to the signal {self._id}")

    def emit(self, *args):
        """
        Emitting a signal will execute all its connected commands
        If an error occurs while executing a command, it will be raised
        """
        for con in self._connections:
            if con.need_arguments:
                con.slot(*args)
            else:
                con.slot()

    def emit_with_catch(self, *args):
        """
        Emitting a signal will execute all its connected commands
        If an error occurs while executing a command, it will be logged
        The error ApplicationExit is not catched
        """

        for con in self._connections:
            try:
                if con.need_arguments:
                    con.slot(*args)
                else:
                    con.slot()
            except ApplicationExit as e:
                raise e
            except Exception as e:
                LOGGER.warning(f"Error : {e} -- while exectuting {con}")


class Connection:

    def __init__(self, owner, signal, slot):

        self.owner = owner
        self.signal = signal
        self.slot = slot
        self.need_arguments = len(inspect.signature(slot).parameters) > 0

    def kill(self):

        self.signal._connections.remove(self)
        if self.owner is not None:
            self.owner._connections.remove(self)


class Communicative(CommunicativeDoc):

    def __init__(self):

        self.signal = Object()
        self._connections = set()

    def create_signal(self, signal_id):

        if not isinstance(signal_id, str) or not (signal_id == signal_id.upper()):
            raise PermissionError("A signal id must be an uppercase str")
        if hasattr(self.signal, signal_id):
            raise PermissionError("A signal already has this id : {}".format(signal_id))

        setattr(self.signal, signal_id, Signal(emitter=self, id=signal_id))

    def disconnect(self):

        for con in tuple(self._connections):
            con.kill()




from baopig import *


class UT_Button_Zone(Zone):
    def __init__(self, *args, **kwargs):
        Zone.__init__(self, *args, **kwargs)

        layer = Layer(self, spacing=10)
        z1 = Zone(self, size=("100%", 60), background_color=(150, 150, 150), spacing=5, padding=5)
        z2 = Zone(self, size=("100%", 100), background_color=(150, 150, 150), spacing=5, padding=5)
        z3 = Zone(self, size=("100%", 500), background_color=(150, 150, 150))
        layer.pack()

        # Z1
        class MyButton(Button):
            def handle_validate(self):
                super().handle_validate()
                t.set_text(f"You clicked on : {self.text}")

        MyButton(z1, "Hello world")
        MyButton(z1, "Dlrow olleh")
        b = Button(z1, "Other way")
        Button(z1, "Console print", command=PrefilledFunction(print, "Hello world from 'Console print' Button"))
        t = Text(z1)
        b.command = PrefilledFunction(t.set_text, f"You clicked on : {b.text}")
        z1.pack(axis="horizontal")

        # Z2
        GridLayer(z2)
        z2.set_style_for(Button, padding=0)
        Button(z2, "l1", row=0, col=0)
        Button(z2, "l1\nl2\nl3", row=0, col=1)
        Button(z2, "-----------l1-----------", row=0, col=2)
        Text(z2, text=f"padding:{z2.get_style_for(Button)['padding']}", row=0, col=3)
        z2.set_style_for(Button, padding=13)
        Button(z2, "l1", row=1, col=0)
        Button(z2, "l1\nl2", row=1, col=1)
        Button(z2, "-----------l1-----------", row=1, col=2)
        Text(z2, text=f"padding:{z2.get_style_for(Button)['padding']}", row=1, col=3)
        z2.pack()

        # Z3
        class PosButton(Button):
            def __init__(self, parent, pos):
                Button.__init__(self, parent, pos=pos, text="({}, {})".format(*pos))

        import random
        width = Button.STYLE["width"]
        height = Button.STYLE["height"]
        for i in range(10):
            x = random.randrange(z3.rect.w - width)
            y = random.randrange(z3.rect.h - height)
            PosButton(z3, pos=(x, y))


# For the PresentationScene import
ut_zone_class = UT_Button_Zone

if __name__ == "__main__":
    from baopig.prefabs.testerscene import TesterScene
    app = Application()
    TesterScene(app, ut_zone_class)
    app.launch()




import glob
import pygame
from baopig.io import LOGGER
# TODO : change the height system ? go back to default from pygame ?


class Font:

    def __init__(self, owner=None, file=None, height=None, color=None, bold=None, italic=None, underline=None):

        self._file = None
        self._filepath = None
        self._height = None
        self._ascent = None
        self._font = None
        self._color = None
        self._owner_wkref = (lambda: None) if owner is None else owner.get_weakref()

        if owner is not None:
            if file is None:
                file = owner.style["font_file"]
            if height is None:
                height = owner.style["font_height"]
            if color is None:
                color = owner.style["font_color"]
            if bold is None:
                bold = owner.style["font_bold"]
            if italic is None:
                italic = owner.style["font_italic"]
            if underline is None:
                underline = owner.style["font_underline"]

        assert file is None or isinstance(file, str)

        self.config(file=file, height=height, color=color, bold=bold, italic=italic, underline=underline,
                    update_owner=False)

    def __str__(self):

        return f"Font(height:{self.height}, ascent:{self._ascent}, file={self.file})"

    __repr__ = __str__

    color = property(lambda self: self._color)
    file = property(lambda self: self._file)
    filepath = property(lambda self: self._filepath)
    height = property(lambda self: self._height)
    owner = property(lambda self: self._owner_wkref())

    def config(self, file=None, height=None, color=None, bold=None, italic=None, underline=None, update_owner=True):

        if file is not None:
            self._file = file

            if file in sys_fonts:
                pass
            elif file == "monospace":
                file = "JetBrainsMono-Medium.ttf"
            elif not file.endswith(".ttf"):
                if file.count("."):
                    LOGGER.warning("Can only look for tff fonts, not " + file)
                    self._file = file = None
                else:
                    file += ".ttf"

            if file in sys_fonts:
                self._filepath = file
            else:
                if file not in filepaths:
                    # print("get_filepath(", file, end=") -> ", sep="")
                    filepath = get_filepath(file)
                    if filepath is None:
                        LOGGER.warning(f"Font not found : {file} (the default font will be used instead)")
                    filepaths[file] = filepath
                    # print(filepaths[file])
                self._filepath = filepaths[file]

        if height is not None:
            assert height > 1, "A font height must be higher than 1"
            self._height = height
        if file is not None or height is not None:
            """if self.height in _all[file]._fonts:
                ascent = _all[file].get_font(self.height)._ascent
            else:
                ascent = self.height
                all_from_file = _all[self.file]

                little_brother = self.height - 1
                if not all_from_file._fonts:
                    little_brother = 0
                while little_brother > 0 and little_brother not in all_from_file._fonts:
                    little_brother -= 1
                if little_brother > 0:
                    ascent = all_from_file.get_font(little_brother)._ascent

                elif all_from_file._fonts and little_brother == 0:
                    big_brother = self.height + 1
                    while big_brother not in all_from_file._fonts:
                        big_brother += 1
                    ascent = all_from_file.get_font(big_brother)._ascent
                    while height < all_from_file.get_font(ascent=ascent).size("")[1]:
                    while height < pygame.font.Font(self.font, ascent).size("")[1]:
                        ascent -= 1
                        if ascent <= 0:
                            raise PermissionError("Too small value for font height : " + str(self.height))

                while self.height < pygame.font.Font(self.file, ascent).size("")[1]:
                    ascent -= 1
                while self.height > pygame.font.Font(self.file, ascent).size("")[1]:
                    ascent += 1
                if ascent <= 0:
                    raise PermissionError("Too small value for font height : " + str(self.height))
                
            self._ascent = ascent
            self._font = pygame.font.Font(self.file, self._ascent)"""

            self._font = _all[self.filepath].get_font(height=self.height)
            self._ascent = self._font.get_ascent()

        if color is not None:
            try:
                color = pygame.Color(color)  # TODO little rework of Color
            except ValueError:
                color = pygame.Color(*color)
            self._color = color
        if bold is not None:
            self._font.set_bold(bold)
        if italic is not None:
            self._font.set_italic(italic)
        if underline is not None:
            self._font.set_underline(underline)

        if update_owner and self.owner is not None:
            self.owner.set_text(self.owner.text)

    def get_width(self, text):

        return self._font.size(text)[0]

    def render(self, text, background_color=None):

        assert background_color is None or len(background_color) == 3, \
            "Don't like rendering a text with a transparent background color"

        rendering = self._font.render(text, 1, self._color, background_color)
        if rendering.get_height() != self._height:
            assert rendering.get_height() > self._height
            rendering = rendering.subsurface((0, 0, rendering.get_width(), self._height)).copy()
        return rendering


class _Fonts:
    """
    Pour plus de clarte, il a ete decide que Font(30) cree une ploice de taille 30
    En effet, pygame.font.Font(30) cree une police de hauteur 21
    Afin de trouver efficacement quelle valeur faut-il donner à pygame.font.Font
    pour obtenir une police de taille 30, on enregistre les polices auparavent crees
    dans _fonts

    Exemple :

    """

    def __init__(self, filepath):

        if filepath == "default":
            filepath = None
        pygame.font.init()
        self._fonts = {}
        self._fonts_by_ascent = {}
        self._filepath = filepath
        if filepath in sys_fonts:
            self._font_class = pygame.font.SysFont
        else:
            self._font_class = pygame.font.Font

    def get_font(self, height=None, ascent=None):
        """
        Methode qui renvoie une police de hauteur 'height' en pixels ou une police cree
        avec la hauteur bizarre de pygame, appelee ici 'ascent'
        """
        if ascent is None:
            assert isinstance(height, int), "height must be an integer (it's a font size)"
            assert height > 0, "A text must have a positive height"

            try:
                # Si le dictionnaire des polices contient deja la taille de police demandee
                return self._fonts[height]
            except KeyError:
                pass

            # Sinon on l'ajoute au dictionnaire
            ascent = height

            while height < self.get_font(ascent=ascent).get_height():
                ascent -= 1
            while height > self.get_font(ascent=ascent).get_height():
                ascent += 1
            if ascent <= 0:
                raise PermissionError("Too small value for font height : " + str(height))

            try:
                # Si le dictionnaire des polices contient deja la taille de police demandee
                return self._fonts[height]
            except KeyError:
                pass

            if self.get_font(ascent=ascent).get_height() < height:
                # print(self._fonts, self.get_font(ascent=ascent).get_height(), height)
                raise AssertionError
            slightly_higher_font = self.get_font(ascent=ascent)
            for i in range(self.get_font(ascent=ascent).get_height() - height):
                self._fonts[height + i + 1] = slightly_higher_font
            self._fonts_by_ascent[ascent] = slightly_higher_font
            return slightly_higher_font

        else:
            assert height is None

            try:
                # Si le dictionnaire des polices contient deja la taille de police demandee
                return self._fonts_by_ascent[ascent]
            except KeyError:
                pass

            new_font = self._font_class(self._filepath, ascent)
            self._fonts[new_font.get_height()] = new_font
            self._fonts_by_ascent[ascent] = new_font
            return new_font


class _All(dict):
    def __getitem__(self, filepath):
        if filepath not in self:
            self[filepath] = _Fonts(filepath)
        return super().__getitem__(filepath)


_all = _All()


def get_filepath(file):
    import os
    exec_dir = os.getcwd()
    for path in glob.iglob(exec_dir + "/*.ttf"):
        if path.endswith(file):
            return path
    for path in glob.iglob(exec_dir + "/*/*.ttf"):
        if path.endswith(file):
            return path
    for path in glob.iglob(exec_dir + "/*/*.ttf"):
        if path.endswith(file):
            return path
    for path in glob.iglob(exec_dir + "/*/*/*.ttf"):
        if path.endswith(file):
            return path
    for path in glob.iglob(__file__[:-7] + "lib/*.ttf"):
        if path.endswith(file):
            return path
    for path in glob.iglob(__file__[:-7] + "lib/*/*.ttf"):
        if path.endswith(file):
            return path
    for path in glob.iglob(__file__[:-7] + "lib/*/*/*.ttf"):
        if path.endswith(file):
            return path
    for path in glob.iglob("/System/Library/Fonts/*.ttf"):
        if path.endswith(file):
            return path
    for path in glob.iglob("/System/Library/Fonts/*.ttf"):
        if path.endswith(file):
            return path
    for path in glob.iglob("/Library/Fonts/Microsoft/*.ttf"):
        if path.endswith(file):
            return path
    for path in glob.iglob("/Library/Fonts Disabled/*.ttf"):
        if path.endswith(file):
            return path
    for path in glob.iglob("/Library/Fonts/*.ttf"):
        if path.endswith(file):
            return path


filepaths = {}
sys_fonts = pygame.font.get_fonts()

# The 2 last font files, on my ACER with Windows 11, are bugged
sys_fonts.remove('leelawadeegras')
sys_fonts.remove('microsoftuighurgras')

if __name__ == "__main__":
    import os

    exec_dir = os.getcwd()
    for path in glob.iglob(__file__[:-7] + "lib/*.ttf"):
        print(path)
    for path in glob.iglob(__file__[:-7] + "lib/*/*.ttf"):
        print(path)
    for path in glob.iglob(__file__[:-7] + "lib/*/*/*.ttf"):
        print(path)
    for path in glob.iglob("/System/Library/Fonts/*.ttf"):
        print(path)
    for path in glob.iglob("/System/Library/Fonts/*.ttf"):
        print(path)
    for path in glob.iglob("/Library/Fonts/Microsoft/*.ttf"):
        print(path)
    for path in glob.iglob("/Library/Fonts Disabled/*.ttf"):
        print(path)
    for path in glob.iglob("/Library/Fonts/*.ttf"):
        print(path)

    h = 50
    font = Font(file='monospace', height=h)
    print(font.height, h, font._font.get_height())
    assert font.height == h == font._font.get_height()
    print(font)




import os
import sys
import logging
import tempfile
import time
from logging import FileHandler

formatters = [logging.Formatter("%(asctime)s :: %(threadName)s :: %(module)s :: %(funcName)s :\n"
                                "    %(levelname)s :: %(msg)s"),  # WARNING, INFO
              logging.Formatter("%(asctime)s :: %(module)s :: %(funcName)s :\n    %(levelname)s :: %(msg)s"),
              logging.Formatter("%(module)s :: %(funcName)s :: %(levelname)s :: %(msg)s"),  # DEBUG
              logging.Formatter("%(asctime)s :: %(levelname)s :: %(msg)s"),
              logging.Formatter("%(levelname)s :: %(msg)s")]  # Console


class PersonalizedLogger(logging.Logger):

    CRITICAL = logging.CRITICAL  # 50
    FATAL = logging.FATAL        # 50
    ERROR = logging.ERROR        # 40
    WARNING = logging.WARNING    # 30
    WARN = logging.WARN          # 30
    INFO = logging.INFO          # 20
    DEBUG = logging.DEBUG        # 10
    FINE = 5
    NOTSET = logging.NOTSET      # 0

    def __init__(self):
        logging.Logger.__init__(self, "LOGGER", level=logging.NOTSET)

        # File Handler
        # Will create the file activity.log in the directory /logs
        # from the directory where the application have been launched
        basename = os.path.splitext(os.path.basename(sys.argv[0]))[0]
        # self.logs_directory = os.path.abspath("logs") + os.path.sep + basename + os.path.sep
        self.logs_directory = tempfile.gettempdir() + "/logs" + os.path.sep + basename + os.path.sep
        os.makedirs(self.logs_directory, exist_ok=True)
        self.open_file_handlers = {}
        for level in (logging.WARNING, logging.INFO):
            self.add_filehandler(level)

        # Console Handler
        self.cons_handler = logging.StreamHandler()
        self.cons_handler.setFormatter(formatters[-1])
        self.cons_handler.setLevel(logging.INFO)
        self.addHandler(self.cons_handler)

        # Done
        self.debug("File and Console logging successfully initialised")

    def add_filehandler(self, level):
        """
        Add to self.handlers a new FileHandler
        His level is set with the argument 'level'
        :param level: a logging existing level
        :return: None
        """
        assert isinstance(level, int), "level must be an integer"

        level_name = logging.getLevelName(level)
        filename = self.logs_directory + level_name + ".log"
        file = open(filename, 'w')
        file.write("\n-------------------------------------------------------------")
        file.write("\n                LOG FILE")
        file.write("\n            {}".format(time.strftime("%b %d, %Y at %H:%M and %S seconds", time.localtime())))
        file.write("\n            File log level:{}".format(level_name))
        file.write("\n            Executing script:{}".format(sys.argv[0]))
        file.write("\n-------------------------------------------------------------\n\n")
        file.close()
        file_handler = FileHandler(filename)
        file_handler.setLevel(level)
        if level != logging.DEBUG:
            file_handler.setFormatter(formatters[0])
        if level == logging.DEBUG:
            file_handler.setFormatter(formatters[2])
        self.addHandler(file_handler)
        self.open_file_handlers[level] = (1, file_handler)

    def add_debug_filehandler(self):
        if logging.DEBUG not in self.open_file_handlers:
            self.add_filehandler(logging.DEBUG)

    def fine(self, msg, *args, **kwargs):
        """
        Log 'msg % args' with severity 'FINE'.

        To pass exception information, use the keyword argument exc_info with
        a true value, e.g.

        logger.fine("Houston, we have a %s", "interesting problem", exc_info=1)
        """
        if self.isEnabledFor(self.FINE):
            self._log(self.FINE, msg, args, **kwargs)


LOGGER = PersonalizedLogger()
# logging.addLevelName(LOGGER.FINE, "FINE")





import pygame
from baopig.lib import Container, paint_lock, Widget


class Zone(Container):

    STYLE = Container.STYLE.substyle()
    STYLE.modify(
        width=100,
        height=100,
    )

    def divide(self, side, width):
        # TODO : rework this function
        if side == "left":
            self.rect.left = width
            self.rect.w -= width
            if self.rect.w <= 0:
                raise ValueError("the new zone shouldn't completly override the central zone")
            zone = Zone((width, self.rect.h))
        else:
            return
        return zone


class SubZone(Zone):  # TODO : SubScene ? with rects_to_update ?
    # TODO : solve : when update on a SubZone whose parent is a scene, the display isn't updated
    """A SuzZone is an optimized Zone, its surface is a subsurface of its parent (cannot have transparency)"""

    def __init__(self, parent, **kwargs):

        Zone.__init__(self, parent, **kwargs)
        try:
            self._surface = self.parent.surface.subsurface(self.rect.topleft + self.rect.size)
        except ValueError:
            assert not self.parent.auto_rect.contains(self.rect)
            raise PermissionError("A SubZone must fit inside its parent")

        self.parent.signal.NEW_SURFACE.connect(self._update_subsurface, owner=self)
        self.signal.MOTION.connect(self._update_subsurface, owner=self)

    def _update_subsurface(self):

        with paint_lock:
            try:
                Widget.set_surface(self, self.parent.surface.subsurface(self.rect.topleft + self.rect.size))
            except ValueError:
                assert not self.parent.auto_rect.contains(self.rect)
                Widget.set_surface(self, self.parent.surface.subsurface(
                    pygame.Rect(self.rect).clip(self.parent.auto_rect)))  # resize the subzone

    def _flip(self):  # TODO : check with new padding & etc
        """Update all the surface"""

        if self.is_hidden:
            return

        with paint_lock:  # TODO : usefull ?

            self._flip_without_update()

            # optimization
            if self.parent is self.scene:
                pygame.display.update(self.hitbox)
            else:
                self.parent.send_display_request(
                    rect=(self.parent.left + self.hitbox.left, self.parent.top + self.hitbox.top) + self.hitbox.size
                )

    def _update_surface_from_resize(self, asked_size):

        with paint_lock:
            try:
                Widget.set_surface(self, self.parent.surface.subsurface(self.rect.topleft + asked_size))
            except ValueError:
                assert not self.parent.auto_rect.contains(self.rect)
                raise PermissionError("A SubZone must fit inside its parent")
            self._flip_without_update()

    def _warn_parent(self, rect):
        """Request updates at rects referenced by self"""

        rect = (self.rect.left + rect[0], self.rect.top + rect[1]) + tuple(rect[2:])

        # because of subsurface, we can skip self.parent._update_rect()
        if self.parent is self.scene:
            pygame.display.update(rect)
        else:
            self.parent._warn_parent(rect)



from baopig import *


class RainbowRect(Rectangle, Runable):
    def __init__(self, parent, **kwargs):
        self.paint_calls = 0
        Rectangle.__init__(self, parent, color=(255, 0, 0), **kwargs)
        Runable.__init__(self, parent)
        self.console = None

    def set_running(self, val):

        if val == self.is_running:
            return

        super().set_running(val)

        if self.is_running:
            self.console.set_text(self.console.text + "\nMethod called : set_running(True)")
        else:
            self.console.set_text(self.console.text + "\nMethod called : set_running(False)")
        self.console.parent.pack(adapt=True)

    def paint(self):
        self.color.set_hue(self.paint_calls)
        super().paint()

    def run(self):
        self.paint_calls = (self.paint_calls + .003) % 360
        self.send_paint_request()


class UT_Runable_Zone(Zone):
    def __init__(self, *args, **kwargs):
        Zone.__init__(self, *args, **kwargs)

        Layer(self, name="zones_layer", spacing=10)
        z1 = Zone(self, background_color=(150, 150, 150), padding=10, spacing=10)
        self.default_layer.pack()

        rainbow = RainbowRect(z1)
        buttons_zone = Zone(z1, spacing=10)
        buttons_zone.set_style_for(Button, width=200)
        Button(buttons_zone, text="start running", command=PrefilledFunction(rainbow.set_running, True))
        Button(buttons_zone, text="stop running", command=PrefilledFunction(rainbow.set_running, False))
        buttons_zone.pack(axis="horizontal", adapt=True)
        rainbow.resize(*buttons_zone.rect.size)
        rainbow.console = Text(z1, text="Console:", max_width=rainbow.rect.w)
        z1.pack(adapt=True)

    def load_sections(self):
        self.parent.add_section(
            title="Paintable.set_running()",
            tests=[
                "After set_running(True), the Runable is running",
                "After set_running(False), the Runable is stopped",
            ]
        )


# For the PresentationScene import
ut_zone_class = UT_Runable_Zone

if __name__ == "__main__":
    from baopig.prefabs.testerscene import TesterScene
    app = Application()
    TesterScene(app, ut_zone_class)
    app.launch()




import pygame
from baopig.lib import Rectangle, Runable


class ProgressBar(Rectangle, Runable):  # TODO : tests & documentation
    STYLE = Rectangle.STYLE.substyle()
    STYLE.modify(
        border_width=2
    )

    def __init__(self, parent, minval, maxval, get_progress, **kwargs):

        try:
            Rectangle.__init__(self, parent, **kwargs)
        except AttributeError:
            pass  # 'ProgressBar' object has no attribute '_progression'
        Runable.__init__(self, parent, **kwargs)

        # Non protected fields (don't need it)
        self._minval = minval
        self._maxval = maxval
        self.get_progress = get_progress
        # Protected field
        self._progression = 0  # percentage between 0 and 1

        self.run()
        self.set_running(True)

    progression = property(lambda self: self._progression)

    def paint(self):
        self.surface.fill((0, 0, 0, 0))
        pygame.draw.rect(self.surface, self.color, (0, 0, self.progression * self.rect.w, self.rect.h))
        pygame.draw.rect(self.surface, self.border_color, self.auto_hitbox, self.border_width * 2 - 1)

    def run(self):
        self._progression = (float(self.get_progress()) - self._minval) / (self._maxval - self._minval)
        self.send_paint_request()



import pygame
from baopig.documentation import Selector as SelectorDoc
from baopig.io import keyboard, mouse
from .utilities import paint_lock
from .widget import Widget
from .widget_supers import Focusable
from .shapes import Rectangle
from .layer import Layer
from .layersmanager import LayersManager
from .container import Container


class SelectableWidget(Widget):
    """
    A SelectableWidget is a Widget who can be selected

    You are selecting a SelectableWidget when :
        - You click on its selector (a parent), and then move the mouse while it is clicked, and
          the rect representing the drag collide with the SelectableWidget
        - You pressed Ctrl+A while its selector is focused

    The selection_rect closes when :
        - A mouse.LEFTCLICK occurs (not double clicks, wheel clicks...)
        - The SelectableWidget size or position changes

    You can rewrite the check_select(abs_rect) method for accurate selections (like SelectableText)
    You can rewrite the select() and unselect() methods for specific behaviors (like SelectableText)

    The get_selected_data() method return, when the SelectableWidget is selected, the SelectableWidget itself
    You will probably want to override this method
    """

    def __init__(self, parent, **kwargs):

        Widget.__init__(self, parent, **kwargs)

        self._is_selected = False

        selector = self.parent
        while not isinstance(selector, Selector):
            selector = selector.parent
        self._selector_ref = selector.get_weakref()

        assert self not in self.selector.selectables
        self.selector.selectables.add(self)

        self.signal.HIDE.connect(self.unselect, owner=None)
        self.signal.SLEEP.connect(self.unselect, owner=None)
        self.signal.KILL.connect(lambda: self.selector.selectables.remove(self), owner=self)

    is_selected = property(lambda self: self._is_selected)
    selector = property(lambda self: self._selector_ref())

    def check_select(self, selection_rect: Rectangle):
        """
        Method called by the selector each time the selection_rect rect changes
        The selection_rect has 3 attributes:
            - start
            - end
            - rect (the surface between start and end)
        These attributes are absolutely referenced, wich means they are relative
        to the application. Start and end attributes reffer to the start and end
        of the selection_rect, who will often be caused by a mouse link motion
        """

        assert self.is_alive
        if self.abs_hitbox.colliderect(selection_rect.abs_rect):
            if not self._is_selected:
                self._is_selected = True
                self.handle_select()
        else:
            self.unselect()

    def get_selected_data(self):

        if self.is_selected:
            return self

    def handle_select(self):
        """
        Called each time the selection rect move and collide with this SelectableWidget
        """

    def handle_unselect(self):
        """
        Called when the selection rect don't collide anymore with this SelectableWidget
        """

    def unselect(self):
        if self.is_selected:
            self._is_selected = False
            self.handle_unselect()


class SelectionRect(Rectangle):
    STYLE = Rectangle.STYLE.substyle()
    STYLE.modify(
        color="theme-color-selection_rect",
        border_color="theme-color-selection_rect",
        border_width=1,
    )

    def __init__(self, parent, abs_start, **kwargs):
        Rectangle.__init__(self, parent=parent, layer=parent.selectionrect_layer,
                           name=parent.name + ".selection_rect", **kwargs)
        self.set_color(tuple(self.color)[:3] + (40,))

        self.abs_start = pygame.Vector2(abs_start)
        self.abs_end = None
        self.set_end(self.abs_start)

        self._can_handle_motion = True
        self.signal.MOTION.connect(self.handle_motion, owner=None)

    def clip(self, rect):
        """
        Clip the rectangle inside another (wich are both relative to the parent)
        """
        new_rect = self.rect.clip(rect)
        if new_rect == self.rect:
            return
        self.set_pos(topleft=new_rect.topleft)
        self.resize(*new_rect.size)

    def handle_motion(self, dx, dy):

        # This is used, for example, when a TextEdit is scrolled while some of its text is selected
        if self._can_handle_motion:
            self.abs_start += (dx, dy)
            self.abs_end += (dx, dy)

    def set_end(self, abs_pos):

        self.abs_end = pygame.Vector2(abs_pos)

        start = self.abs_start - self.parent.abs_rect.topleft
        end = self.abs_end - self.parent.abs_rect.topleft
        rect = pygame.Rect(start, end - start)
        rect.normalize()

        self._can_handle_motion = False
        self.set_pos(topleft=rect.topleft)
        self.resize(width=rect.w + 1, height=rect.h + 1)  # the selection_rect rect collide with mouse.pos
        self.clip(self.parent.auto_rect)
        self._can_handle_motion = True

    def set_visibility(self, visible):

        if visible:
            self.show()
        else:
            self.hide()


class Selector(SelectorDoc, Container, Focusable):

    def __init__(self, parent, can_select=True, selectionrect_class=SelectionRect, **kwargs):

        Container.__init__(self, parent, **kwargs)
        Focusable.__init__(self, parent)

        self.selectables = set()
        self._can_select = can_select
        self._selectionrect_class = selectionrect_class
        self._selection_rect_ref = lambda: None
        self._selectionrect_visibility = True
        self.selectionrect_layer = Layer(self, self._selectionrect_class, level=LayersManager.FOREGROUND)

    def _get_iselected(self):
        for widget in self.selectables:
            if widget.is_selected:
                yield widget

    _iselected = property(_get_iselected)
    can_select = property(lambda self: self._can_select)
    is_selecting = property(lambda self: self._selection_rect_ref() is not None)
    selection_rect = property(lambda self: self._selection_rect_ref())

    def del_selection_data(self):
        """This method is called when the user press Ctrl + X"""

    def close_selection(self):

        if not self.is_selecting:
            return
        self.selection_rect.kill()
        for selectable in self._iselected:
            if not selectable.is_selected:
                continue
            selectable._is_selected = False
            selectable.handle_unselect()

    def end_selection(self, abs_pos, visible=None):
        """
        Sets up the selection_rect & checks the selectables which are selected
        :param abs_pos: An absolute position -> relative to the scene
        :param visible: If you want to change the visibility until the next end_selection
        """

        if not self._can_select:
            return
        assert self.selection_rect is not None
        if abs_pos == self.selection_rect.abs_end:
            return
        if visible is not None:
            self.selection_rect.set_visibility(visible)
        else:
            self.selection_rect.set_visibility(self._selectionrect_visibility)
        self.selection_rect.set_end(abs_pos)
        for selectable in self.selectables:
            selectable.check_select(self.selection_rect)

    def get_selection_data(self):

        if not self.is_selecting:
            return
        selected = tuple(self._iselected)
        if not selected:
            return  # happens when the selection_rect didn't select anything
        sorted_selected = sorted(selected, key=lambda o: (o.abs_rect.top, o.abs_rect.left))
        return '\n'.join(str(s.get_selected_data()) for s in sorted_selected)

    def handle_defocus(self):

        super().handle_defocus()
        self.close_selection()

    def handle_keydown(self, key):

        super().handle_keydown(key)  # TAB and arrows management

        if keyboard.mod.ctrl:

            if key == pygame.K_a:  # Ctrl + a -> select all
                self.select_all()

            elif key == pygame.K_c:  # Cmd + c -> copy selected data
                selected_data = self.get_selection_data()
                if selected_data:
                    pygame.scrap.put(pygame.SCRAP_TEXT, str.encode(selected_data))

            elif key == pygame.K_v:  # Cmd + v -> paste clipboard data
                scrap_bytes = pygame.scrap.get(pygame.SCRAP_TEXT)
                if scrap_bytes is not None:
                    text = scrap_bytes.decode()
                    text = str.replace(text, '\0', '')  # removes null characters
                    text = str.replace(text, '\r', '')  # removes carriage return characters
                    self.paste(text)

            elif key == pygame.K_x:  # Cmd + x -> copy and cut selected data
                selected_data = self.get_selection_data()
                if selected_data:
                    pygame.scrap.put(pygame.SCRAP_TEXT, str.encode(selected_data))
                    self.del_selection_data()

    def handle_link(self):

        self.close_selection()  # only usefull at link while already focused
        for s in self.selectables:
            if hasattr(s, "handle_selector_link"):
                s.handle_selector_link()  # double & triple clicks on text

    def handle_link_motion(self, rel):
        with paint_lock:
            if self.selection_rect is None:
                link_origin = mouse.x - rel[0], mouse.y - rel[1]
                self.start_selection(link_origin)
            self.end_selection(mouse.pos)

    def handle_unlink(self):

        if self.selection_rect:
            self.selection_rect.hide()

    def paste(self, data):
        """This method is called when the user press Ctrl + V"""

    def select_all(self):  # TODO : solve : selects also hidden selectables

        if self.selectables:
            if self.is_selecting:
                self.close_selection()
            self.start_selection(self.abs_rect.topleft)
            self.end_selection(self.abs_rect.bottomright, visible=False)

    def set_selectionrect_visibility(self, visible):

        self._selectionrect_visibility = bool(visible)
        if self.selection_rect is not None:
            self.selection_rect.set_visibility(visible)

    def start_selection(self, abs_pos):
        """
        A selection_rect can only be started once
        :param abs_pos: An absolute position -> relative to the scene
        """

        if not self._can_select:
            return
        if self.selection_rect is not None:
            raise PermissionError("A selection must be closed before creating a new one")

        self._selection_rect_ref = self._selectionrect_class(
            self, abs_start=abs_pos, visible=self._selectionrect_visibility
        ).get_weakref()



import time
from baopig.time.stopwatch import _running_stopwatches
from baopig.time.timer import _running_timers, timer_lock, RepeatingTimer


# TODO : baopig.time.init() for chifoumi__server


class _TimeManager:

    def __init__(self):

        self._paused_timers = None
        self._paused_stopwatches = None

    """def pause_TBR(self):

        if self._paused_timers is not None:
            raise PermissionError("The TimeThread is already paused")

        self._paused_timers = tuple(_running_timers)
        for timer in self._paused_timers:
            timer.pause()
        self._paused_stopwatches = tuple(_running_stopwatches)
        for stopwatch in self._paused_stopwatches:
            stopwatch.stop()
        self._is_paused = True

    def resume_TBR(self):

        if self._paused_timers is None:
            raise PermissionError("The TimeThread is already running")

        for timer in self._paused_timers:
            timer.resume()
        self._paused_timers = None
        for stopwatch in self._paused_stopwatches:
            stopwatch.start()
        self._paused_stopwatches = None"""

    @staticmethod
    def update():

        current_time = time.time()

        with timer_lock:
            for timer in tuple(_running_timers):

                if timer._end_time is None:  # happens when timer.TIMEOUT provokes another_timer.cancel()
                    continue

                if timer._end_time <= current_time:
                    if isinstance(timer, RepeatingTimer):
                        timer._repeat()
                    else:
                        _running_timers.remove(timer)
                    timer.signal.TIMEOUT.emit()


time_manager = _TimeManager()
_running_timers.manager = time_manager
_running_stopwatches.manager = time_manager
del _TimeManager



from baopig.io import LOGGER
from .utilities import *
from .style import Theme
from .layer import Layer
from .zone import Zone, Widget
from .selections import Selector


class Scene(Zone, Selector, Handler_SceneOpen, Handler_SceneClose):
    """
    A Scene is like a page in an application. It can be the menu, the parameters page...

    If size parameter is filled, then when we swap to this scene, the application will be resized
    to 'size'. When it will swap again to another scene, if no size is required, the application
    will be resized to its original size.
    """

    STYLE = Zone.STYLE.substyle()
    STYLE.modify(
        background_color="theme-color-scene_background",
    )

    def __init__(self, application, size=None, can_select=True, **kwargs):

        if "name" not in kwargs:
            kwargs["name"] = self.__class__.__name__
        self._application = application

        if "theme" in kwargs:
            theme = kwargs.pop("theme")
            if not isinstance(theme, str):  # theme name
                assert isinstance(theme, Theme)
                # if not theme.issubtheme(application.theme):
                #     raise PermissionError("Must be an application sub-theme")
        else:
            theme = application.theme.subtheme()

        self._theme = theme
        Zone.__init__(self, parent=self, pos=(0, 0), size=application.default_size if size is None else size,
                      **kwargs)
        Selector.__init__(self, parent=self, can_select=can_select)

        # self._mode = 0
        self._size_is_shared_with_appliation = size is None
        self._asked_size = self.rect.size
        self._mode_before_fullscreen = None
        self._size_before_fullscreen = None
        self._focused_widget_ref = lambda: None

    def __str__(self):

        return self.name

    asked_size = property(lambda self: self._asked_size)
    focused_widget = property(lambda self: self._focused_widget_ref())
    # mode = property(lambda self: self._mode)
    painter = property(lambda self: self.application._painter)
    scene = property(lambda self: self)  # End of recursive call
    size = property(lambda self: self._rect.size)

    def _add_child(self, widget):

        if widget is self:
            self.application._add_scene(self)  # a scene is a root
        else:
            super()._add_child(widget)

    def _close(self):

        if self.application.focused_scene is not self:
            return
        self._container_close()
        self.handle_scene_close()
        Widget.set_surface(self, pygame.Surface(self.rect.size))  # not pygame.display anymore
        self.focus(None)
        self.application._focused_scene = None

        # LOGGER.debug("Close scene : {}".format(self))

    def _warn_parent(self, rect):

        pygame.display.update(rect)

        if self.painter.is_recording and self.painter.is_recording.only_at_change:
            pygame.image.save(self.surface,
                              self.painter.out_directory + "record_{:0>3}.png".format(
                                  self.painter.record_index))
            self.painter.record_index += 1

    def divide(self, side, width):
        raise PermissionError("Cannot divide a Scene")  # TODO : rework Zone.divide

    def focus(self, widget):

        if widget == self.focused_widget:
            return

        # Defocus
        old_focused = self.focused_widget
        if old_focused is not None:
            assert old_focused.is_focused
            old_focused._is_focused = False

        if widget is None:
            widget = self
        else:
            assert widget.is_visible
        assert not widget.is_focused

        # Focus
        widget._is_focused = True
        self._focused_widget_ref = widget.get_weakref()  # (lambda: None) if widget is None else

        if old_focused is not None:
            # old_focused.signal.DEFOCUS_.emit()
            old_focused.handle_defocus()
        # widget.signal.FOCUS_.emit()
        widget.handle_focus()

    def handle_event(self, event):
        """ Called at every pygame event, while this scene is open """

    def kill(self):

        if not self.is_alive:
            return
        if self.application.focused_scene is self:
            raise PermissionError("Cannot kill a focused scene")

        with paint_lock:
            for child in tuple(self.children):
                child.kill()
            self.disconnect()
            self.signal.KILL.emit(self._weakref)
            self._weakref._ref = None
            self.application.scenes.remove(self)

        del self

    def open(self):

        app = self.application
        if app.focused_scene is self:
            return

        if self._size_is_shared_with_appliation:
            self.resize(*self.application.default_size)

        with paint_lock:
            scene_to_close = app.focused_scene
            app._focused_scene = self
            if scene_to_close:
                scene_to_close._close()
            app._update_display()
            self._container_open()
            self.handle_scene_open()
            self._container_refresh(recursive=True)

        LOGGER.debug("Open scene : {}".format(self))

    def pre_open(self):
        """Stuff to do right before this scene is open"""

    def _update_surface_from_resize(self, asked_size):

        self.application._update_display()

    def run(self):
        """Stuff to repeat endlessly while this scene is focused"""

    """def set_mode_TBR(self, mode):

        if mode is self.mode:
            return

        assert mode in (0, pygame.NOFRAME, pygame.RESIZABLE, pygame.FULLSCREEN)

        if mode is pygame.FULLSCREEN and self.mode != mode:
            self._mode_before_fullscreen = self.mode
            self._size_before_fullscreen = self.asked_size

            # print("Asked for fullscreen")
            # mode = 0

        self._mode = mode
        self.application._update_display()"""

    def toggle_debugging(self):

        if not hasattr(self, "debug_layer"):
            self.debug_layer = Layer(self, name="debug_layer", level=self.layers_manager.FOREGROUND)
            from baopig.prefabs.debugzone import DebugZone
            self.debug_zone = DebugZone(self)
        else:
            self.debug_zone.toggle_debugging()

    # TODO : fullscreen
    """def toggle_fullscreen(self):  

        if self.mode == pygame.FULLSCREEN:
            self.set_mode(self._mode_before_fullscreen)
        else:
            self.set_mode(pygame.FULLSCREEN)"""

    # FORBIDDEN METHODS
    def sleep(self):
        raise PermissionError("A Scene cannot sleep")

    def hide(self):
        raise PermissionError("A Scene cannot be hidden")

    def set_touchable_by_mouse(self, val):
        raise PermissionError("A Scene has to be touchable")




"""
Welcome to BaoPig

BAOPIG : Boite A Outils Pour Interfaces Graphiques

"""

print("baopig from WIP")

# TODO : compilation executable
from .version.version import version
print(f"baopig {version} - ", end="")
from pygame import *
from .pybao.issomething import *
from .pybao.objectutilities import Object, PrefilledFunction, PackedFunctions

from .io import *
from .time import *
from .lib import *
from .widgets import *

display = None  # protection for pygame.display

__version__ = str(version)


def debug_with_logging():

    LOGGER.add_debug_filehandler()
    LOGGER.cons_handler.setLevel(LOGGER.DEBUG)





from baopig import *
from baopig.version.version import version
from baopig.prefabs.testerscene import TesterScene


class PresentationScene(Scene):

    def __init__(self, app):

        Scene.__init__(self, app)

        app.set_debug(averagefps=True, launchtime=True)

        Text(self, text="")
        Text(self, text=f"Welcome to baopig version {version}")
        Text(self, text="You can look for the tutorial or experiment unit tests")
        Button(self, text="Unit Tests", command=PrefilledFunction(app.open, "UTMenu_Scene"),
               row=len(self.default_layer))
        Button(self, text="Tutorial")
        self.pack()
        UTMenu_Scene(app)
        self.open()  # without this line, UTMenu_Scene is the first scene, since this one has not been added yet


class UTMenu_Scene(Scene):

    def __init__(self, app):

        Scene.__init__(self, app)

        GridLayer(self)

        Text(self, text="", row=0)
        Text(self, text="Which class do you want to test ?", row=1)
        Text(self, text="", row=2)

        back = Button(self, "Menu", command=PrefilledFunction(app.open, "PresentationScene"), col=1)
        back.set_pos(bottomright=self.rect.bottomright)

        def get_ut_filenames():
            import os
            directory = os.path.dirname(os.path.realpath(__file__))[:-7] + "unit_tests"
            for root, dirs, files in os.walk(directory):
                for file_name in files:
                    if file_name.endswith(".py") and file_name.startswith("ut_"):
                        yield file_name[:-3]  # discard '.py'

        import importlib
        for filename in get_ut_filenames():
            if filename == "testerscene":
                continue
            ut_file = importlib.import_module("baopig.unit_tests." + filename)
            try:
                zone_class = ut_file.ut_zone_class

                def open_testerscene(zc):
                    TesterScene(app, zc).open()
                Button(self, row=len(self.default_layer), text=zone_class.__name__[3:-5],  # discards 'UT_' and '_Zone'
                       command=PrefilledFunction(open_testerscene, zone_class), catching_errors=False)  # TODO : True
            except AttributeError:
                pass




from baopig import *


class UT_TextEdit_Zone(Zone):

    def __init__(self, *args, **kwargs):

        Zone.__init__(self, *args, **kwargs)

        z1 = Zone(self, pos=(10, 10), size=(int(self.rect.w / 2) - 15, 150),
                  background_color=(100, 100, 100, 50))
        z2 = Zone(self, pos=(z1.rect.right + 10, 10), size=(int(self.rect.w / 2) - 15, 150),
                  background_color=(100, 100, 100, 50))
        z3 = Zone(self, pos=(10, z1.rect.bottom + 10), size=(int(self.rect.w / 3) - 20, 120),
                  background_color=(100, 100, 100, 50), padding=5, spacing=2)
        z4 = Zone(self, pos=(z3.rect.right + 10, z1.rect.bottom + 10), size=(int(self.rect.w / 3) - 20, 120),
                  background_color=(100, 100, 100, 50), padding=5, spacing=2)

        z1.mirrored = TextEdit(z1, text="0123456789012345", max_width=40, pos=(10, 10), size=(100, 120))
        z1.d = DynamicText(z1, z1.mirrored.text_widget.get_text, pos=z1.mirrored.rect.topright)

        text = TextEdit(z2, pos=(10, 10), size=(z2.rect.w - 20, 120))

        Button(z2, text="RUN", sticky="topright", ref=text, command=lambda: exec(text.text), catching_errors=True)

        # Z3
        LineEdit(z3)
        LineEdit(z3, width=190, padding=5, font_height=30)
        z3.pack()

        # Z4
        Entry(z4, entry_type=int)
        Entry(z4, entry_type=float)
        Entry(z4, entry_type=str)
        NumEntry(z4, minval=0, maxval=10, accept_floats=True, default=5)
        z4.pack()


# For the PresentationScene import
ut_zone_class = UT_TextEdit_Zone

if __name__ == "__main__":
    from baopig.prefabs.testerscene import TesterScene
    app = Application()
    TesterScene(app, ut_zone_class)
    app.launch()



from math import inf as math_inf
from baopig.io import mouse
from baopig.font.font import Font
from baopig.lib import *

separators = " ,„.…:;/\'\"`´”’" \
             "=≈≠+-±–*%‰÷∞√∫" \
             "()[]{}<>≤≥«»" \
             "?¿!¡@©®ª#§&°" \
             "‹◊†¬•¶|^¨~" \
             ""


class _Line(Widget):
    """
    A Line is a component who only have text on its surface
    It has a transparent background
    It has an end string who is the separator between this line and the next one

    The aspects who might evolve :
        - the text
        - the font size
        - the font color
        - the end character

    When the hitbox size change, there is only one point who is used to locate the LineText : it is
    the location. Exemple : if the location is 'topleft', and the hitbox's width grows, it will expend on the
    right side, because the location is on the left. There is 9 possible values for the location :
        - topleft       - midtop        - topright
        - midleft       - center        - midright
        - bottomleft    - midbottom     - bottomright
    The default location is 'topleft'


    Here is an example of font size :

        font_size = 63
        ____   _
        ____  |_   _   _         _             | 13 px  |
        ____  |   |_| |     |_| |_| |_|        | 37 px  | 63 px
        ____                 _|                | 13 px  |

    A Line is an element of a Paragraph, wich ends with a '\n'.

    end :
        '\n' : end of paragraph
        '' : line cutted

    """

    def __init__(self, parent, text, line_index, end):

        assert isinstance(parent, Text)

        font_render = parent.font.render(text)
        surf_w = font_render.get_width()
        surface = pygame.Surface((surf_w, parent.font.height), pygame.SRCALPHA)
        surface.blit(font_render, (0, 0))

        self._line_index = line_index
        Widget.__init__(self, parent=parent, layer=parent.lines, name=f"{self.__class__.__name__[1:]}({text})",
                        surface=surface)

        assert end in ('\n', '')
        if parent.width_is_adaptable:
            assert end == '\n'
        self._end = end

        assert '\n' not in text
        assert '\v' not in text
        text = str.replace(text, '\t', '    ')  # We can also replace it at rendering
        self._text = text

        # char_pos[i] est la distance entre left et la fin du i-eme caractere
        # Exemple : soit self.text = "Hello world"
        #           char_pos[6] = margin + distance entre le debut de "H" et la fin de "w"
        self._chars_pos = []
        self.update_char_pos()

    def __repr__(self):
        return f"{self.__class__.__name__}(index={self.line_index}, text={self.text})"

    def __str__(self):
        return self.text

    end = property(lambda self: self._end)
    font = property(lambda self: self._parent._font)
    line_index = property(lambda self: self._line_index)
    text = property(lambda self: self._text)
    text_with_end = property(lambda self: self._text + self._end)

    def find_index(self, x, only_left=False):
        """
        Renvoie l'index correspondant a la separation de deux lettres la plus proche de x

        Example :
            x = 23              (23eme pixel a droite de self.left)
            find_index(x) -> 3  (position entre la 3eme et la 4eme lettre)

        Si only_left est demande, renvoie un index qui correspond a une separation a droite
        de x
        """

        def ecart(x1, x2):
            return abs(x1 - x2)

        dist_from_closest_char = math_inf
        index_of_closest_char = None

        for index, char_pos in enumerate(self._chars_pos):

            if only_left and char_pos > x:
                break
            if ecart(x, char_pos) > dist_from_closest_char:
                break

            dist_from_closest_char = ecart(x, char_pos)
            index_of_closest_char = index

        return index_of_closest_char

    def find_mouse_index(self):
        """
        Return the closest index from mouse.x
        """

        return self.find_index(mouse.get_pos_relative_to(self)[0])

    def find_pixel(self, index):
        """
        Renvoi la distance entre hitbox.left et la fin du index-eme caractere
        """
        return self._chars_pos[index]

    def get_iparagraph(self):

        if self.line_index == 0:
            first_line_of_paragraph = self
        else:
            i = self.line_index
            while self.parent.lines[i - 1].end != '\n':
                i -= 1
            first_line_of_paragraph = self.parent.lines[i]

        line = first_line_of_paragraph
        while line.end != '\n':
            yield line
            line = self.parent.lines[line.line_index + 1]
        yield line

    def insert(self, char_index, string):
        """
        Insert a string inside a line (not after the end delimitation)
        """

        if string:
            self.config(text=self.text[:char_index] + string + self.text[char_index:])

    def pop(self, index):
        """
        Remove one character from line.text_with_end
        """

        if index < 0:
            index = len(self.text_with_end) + index
        if self.end != '' and index == len(self.text_with_end) - 1:
            if self.line_index == len(self.parent.lines) - 1:
                return  # pop of end of text
            self.config(self.text, end='')
        else:
            self.config(text=self.text[:index] + self.text[index + 1:])

    def config(self, text, end=None):

        with paint_lock:

            assert '\v' not in text
            text = str.replace(text, '\t', '    ')  # We can also replace it at rendering
            self._text = text

            if len(tuple(self.get_iparagraph())) > 0:

                if end is not None:
                    assert end in ('\n', '')
                    if end != '\n' and self.line_index == len(self.parent.lines) - 1:
                        raise AssertionError
                    self._end = end

                text = ''.join(line.text_with_end for line in self.get_iparagraph())[:-1]  # discard '\n'

                for line in tuple(self.get_iparagraph()):
                    if line != self:
                        line.kill()

            line_index = self.line_index
            parent = self.parent
            self.kill()

            for paragraph_text in text.split("\n"):

                create_line = True
                while create_line:

                    line_text, other_lines_text = parent._cut_text(paragraph_text)

                    if other_lines_text == '':
                        end = '\n'
                        create_line = False
                    else:
                        end = ''
                        paragraph_text = other_lines_text

                    new_line = self.__class__(
                        parent=parent,
                        text=line_text,
                        line_index=line_index,
                        end=end,
                    )
                    line_index = new_line.line_index + .5

    def update_char_pos(self):
        """
        Actualise les valeurs de char_pos

        Appele lors de la creation de LineSelection et lorsque le texte change

        char_pos[i] est la distance entre hitbox.left et la fin du i-eme caractere
        Exemple :
                    Soit self.text = "Hello world"
                    char_pos[6] = margin + distance entre le debut de "H" et la fin de "w"
        """
        self._chars_pos = [0]
        text = ''
        for char in self.text:
            text += char
            self._chars_pos.append(self.font.get_width(text))


class _SelectableLine(_Line):
    """
    You are selecting a SelectableLine when :
        - A condition described in SelectableWidget is verified
        - A cursor moves while Maj key is pressed
    """

    # NOTE : this is not beautiful
    _selection_ref = lambda self: None  # needed during construction
    _is_selected = False

    # def __init__TBR(self, *args, **kwargs):
    #     self._selection_ref = lambda: None  # needed during construction
    #     _Line.__init__(self, *args, **kwargs)
    #     self._is_selected = False

    is_selected = property(lambda self: self._is_selected)
    selection = property(lambda self: self._selection_ref())
    selector = property(lambda self: self._parent.selector)

    def check_select(self, selection_rect):
        """
        Method called by the selector each time the selection_rect rect changes
        The selection_rect has 3 attributes:
            - start
            - end
            - rect (the surface between start and end)
        These attributes are absolutely referenced, wich means they are relative
        to the application. Start and end attributes reffer to the start and end
        of the selection_rect, who will often be caused by a mouse link motion
        """

        assert self.is_alive
        collide_rect = (self.parent.abs_rect.left, self.abs_rect.top, self.parent.abs_rect.w, self.abs_rect.h)

        if selection_rect.abs_rect.colliderect(collide_rect):
            self._is_selected = True
            self.handle_select()
        else:
            if not self.is_selected:
                return
            self._is_selected = False
            self.handle_unselect()

    def get_selected_data(self):
        if self.selection is None:
            return ''
        return self.selection.get_data()

    def handle_select(self):

        selection = self.selector.selection_rect
        if self.selection is None and not selection.rect.w and not selection.rect.h:
            return

        if self.selection is None:
            self._selection_ref = _LineSelection(self).get_weakref()  # TODO : solve bug : multiple lines selection

        selecting_line_end = False
        if self.abs_rect.top <= selection.abs_start[1] < self.abs_rect.bottom:
            start = self.find_index(selection.abs_start[0] - self.abs_rect.left)
        elif selection.abs_start[1] < self.abs_rect.top:
            start = 0
        else:
            start = len(self.text)
            if self is not self.parent.lines[-1]:
                selecting_line_end = True

        if self.abs_rect.top <= selection.abs_end[1] < self.abs_rect.bottom:
            end = self.find_index(selection.abs_end[0] - self.abs_rect.left)
        elif selection.abs_end[1] < self.abs_rect.top:
            end = 0
        else:
            end = len(self.text)
            if self is not self.parent.lines[-1]:
                selecting_line_end = True

        start, end = sorted((start, end))
        self.selection.config(start, end, selecting_line_end)

    def handle_unselect(self):
        if self.selection is not None:
            self.selection.kill()

    def select_word(self, index):
        """
        Selectionne le mot le plus proche de index
        """

        index_start = index_end = index
        while index_start > 0 and self.text[index_start - 1] not in separators:
            index_start -= 1
        while index_end < len(self.text) and self.text[index_end] not in separators:
            index_end += 1

        # Si index n'est pas sur un mot, on selectionne tout le label
        if index_start == index_end == index:
            index_start = 0
            index_end = len(self.text)

        with paint_lock:
            if self.selector.is_selecting:
                self.selector.close_selection()
            if index_start == 0:
                self.selector.start_selection((self.abs_rect.left, self.abs_rect.top))
            else:
                self.selector.start_selection((self.abs_rect.left + self.find_pixel(index_start), self.abs_rect.top))
            if index_end == len(self.text):
                self.selector.end_selection((self.abs_rect.right, self.abs_rect.top), visible=False)
            else:
                self.selector.end_selection((self.abs_rect.left + self.find_pixel(index_end), self.abs_rect.top),
                                            visible=False)


class _LineSelection(Rectangle):
    """
    A LineSelection is a Rectangle with a light blue color : (167, 213, 255, 127)
    Each Line can have a LineSelection

    When you click on a SelectableLine, and then move the mouse while its pressed,
    you are selecting the SelectableLine
    The size and the position of the LineSelection object change according to your mouse

    When you double-click on a SelectableLine, it selects a word
    When you triple-click on a SelectableLine, it selects the whole line text
    """

    STYLE = Rectangle.STYLE.substyle()
    STYLE.modify(
        color="theme-color-selection",
        border_width=0
    )

    def __init__(self, line):

        assert isinstance(line, _SelectableLine)

        self._line_index = line.line_index
        Rectangle.__init__(self,
                           parent=line.parent,
                           pos=line.rect.topleft,
                           size=(0, line.rect.h),
                           name=line.name + " -> selection",
                           layer=line.parent.line_selections
                           )

        # self.is_selecting = False  # True if the user is pressing the mouse button for a selection
        self._index_start = self._index_end = 0
        self._is_selecting_line_end = False
        self._line_ref = line.get_weakref()

        # Initializations
        # self.move_behind(self.line)
        self.line.signal.KILL.connect(self.kill, owner=self)

    index_end = property(lambda self: self._index_end)
    index_start = property(lambda self: self._index_start)
    line = property(lambda self: self._line_ref())
    line_index = property(lambda self: self._line_index)
    text = property(lambda self: self.line._text)

    def config(self, index_start, index_end, selecting_line_end):

        index_start, index_end = sorted((index_start, index_end))
        self.set_start(index_start)
        self.set_end(index_end, selecting_line_end)

    def get_data(self):
        end = self.line.end if self._is_selecting_line_end else ''
        if end == '':
            end = ''
        return self.line.text[self.index_start:self.index_end] + end

    def set_end(self, index, selecting_line_end):

        self._index_end = index
        self._is_selecting_line_end = selecting_line_end
        if selecting_line_end:
            assert self.line is not self.parent.lines[-1]

        if self._is_selecting_line_end:
            self.resize_width(self.line.parent.content_rect.width -  # all the way long, further than line.w
                              self.line.find_pixel(self.index_start))
        else:
            self.resize_width(abs(self.line.find_pixel(self.index_end) -
                                  self.line.find_pixel(self.index_start)))
        self.set_pos(left=self.line.rect.x + self.line.find_pixel(self.index_start))

    def set_start(self, index):

        if index == self.index_start:
            return

        self._index_start = self._index_end = index
        self.resize_width(0)
        self.set_pos(left=self.line.find_pixel(self._index_start))

        if self.is_asleep:
            self.wake()
        self.show()


class Text(Zone, SelectableWidget):
    STYLE = Zone.STYLE.substyle()
    STYLE.create(
        align_mode="left",
        font_file=None,
        font_height=15,
        font_color="theme-color-font",
        font_bold=False,
        font_italic=False,
        font_underline=False,
        max_width=None,
        selectable=True,
    )
    STYLE.set_type("align_mode", str)
    STYLE.set_type("font_height", int)
    STYLE.set_type("font_color", Color)
    STYLE.set_type("font_bold", bool)
    STYLE.set_type("font_italic", bool)
    STYLE.set_type("font_underline", bool)
    STYLE.set_type("selectable", bool)
    STYLE.set_constraint("align_mode", lambda val: val in ("left", "center", "right"),
                         "must be 'left', 'center' or 'right'")
    STYLE.set_constraint("font_height", lambda val: val > 0, "a text must have a positive font height")
    STYLE.set_constraint("font_file", lambda val: (val is None) or isinstance(val, str), "must be None or a string")
    STYLE.set_constraint("max_width", lambda val: (val is None) or isinstance(val, int), "must be None or an integer")

    def __init__(self, parent, text="", **kwargs):

        if "width" in kwargs:
            raise PermissionError
        if "height" in kwargs:
            raise PermissionError

        Zone.__init__(self, parent, **kwargs)
        SelectableWidget.__init__(self, parent)

        self._font = Font(self)
        self._lines_pos = []
        self._align_mode = self.style["align_mode"]
        self._is_selectable = self.style["selectable"]
        self._max_width = self.style["max_width"]
        self._padding = self.style["padding"]
        self._has_locked.text = False  # TODO : remove ?

        self.signal.NEW_SURFACE.connect(self.unselect, owner=None)

        self.line_selections = Layer(self, _LineSelection, touchable=False, sort_by_pos=True)
        self.lines = Layer(self, _Line, name="lines", default_sortkey=lambda line: line.line_index)
        self.set_text(text)

        if self._max_width is not None:
            assert self.content_rect.width == self.max_width

    align_mode = property(lambda self: self._align_mode)
    font = property(lambda self: self._font)
    is_selectable = property(lambda self: self._is_selectable)
    max_width = property(lambda self: self._max_width)
    padding = property(lambda self: self._padding)
    width_is_adaptable = property(lambda self: self._max_width is None)

    def set_max_width_TBR(self, max_width):
        """
        Example:
            widget = Text(parent, "Hello world", font_file=None)
            print(widget.size)  # -> prints (82, 15)
            widget.set_max_width(80)
            widget.set_text("Hello world and everyone else")  # -> here, the font_height will be reduced
        """

        if max_width == self._max_width:
            return
        self._max_width = max_width
        self.set_text(self.get_text())

    def _add_child(self, child):

        super()._add_child(child)
        if isinstance(child, _Line):
            self._pack()

    def _cut_text(self, text):

        if self.width_is_adaptable:
            return text, ''

        max_width = self._max_width

        def is_too_long(text):
            return self.font.get_width(text) > max_width

        def get_cut_index(text):
            len_full_text = len(text)
            reversed_full_text = reversed(text)
            last_char_was_a_separator = False
            for char_index, char in enumerate(reversed_full_text):
                if char in separators:
                    yield len_full_text - char_index
                    last_char_was_a_separator = True
                elif last_char_was_a_separator:
                    yield len_full_text - char_index
                    last_char_was_a_separator = False

        if not is_too_long(text):
            return text, ''

        if is_too_long(text[0]):
            # return text[0], text[1:]
            raise ValueError(f"The availible width is too little : {max_width}")

        cut_index = len(text)
        first_part = text
        for cut_index in get_cut_index(text):

            first_part = text[:cut_index]
            if not is_too_long(first_part):
                return first_part, text[cut_index:]

        # Here, there is no separator in first_part, but it is too long

        while is_too_long(first_part):
            cut_index -= 1
            first_part = text[:cut_index]

        return first_part, text[cut_index:]

    def _pack(self):

        centerx = None  # warning shut down
        if self.align_mode == "center":  # only usefull for the widget creation
            if self.width_is_adaptable:
                centerx = max(line.rect.w for line in self.lines) / 2 + self.content_rect.left
            else:
                centerx = self.content_rect.centerx

        self.lines.sort()
        # TODO : test with windowed text aligned to the center, content_rect might not be a good solution
        h = self.content_rect.top
        for i, line in enumerate(self.lines):
            line._line_index = i
            line.set_pos(top=h)
            h = line.rect.bottom
            if self.align_mode == "left":
                line.set_pos(left=self.content_rect.left)
            elif self.align_mode == "center":
                line.set_pos(centerx=centerx)
            elif self.align_mode == "right":
                line.set_pos(right=self.content_rect.right)

        # Adaptable resize
        if self.width_is_adaptable:
            right = max(line.rect.right for line in self.lines)
            bottom = self.lines[-1].rect.bottom
            self.resize(width=right + self.padding.right, height=bottom + self.padding.bottom)
        else:
            bottom = self.lines[-1].rect.bottom
            if bottom + self.padding.bottom != self.rect.h:  # TODO : without this line, the printing bug, find why
                self.resize(width=self._max_width + self.padding.left + self.padding.right,
                            height=bottom + self.padding.bottom)

        # New positions in _lines_pos
        self._lines_pos = []
        for line in self.lines:
            self._lines_pos.append(line.rect.top)

    def _find_index(self, pos):
        """
        Renvoie l'index correspondant a l'espace entre deux caracteres le plus proche

        Exemple :
            pos = (40, 23)                             (40eme pixel a partir de self.left, 23eme pixel sous self.top)
            find_index(pos) -> self.find_index(2, 13)  (pos est sur le 14eme caractere de la 3eme ligne)
        """

        if pos[1] < 0:
            return self.lines[0].find_index(pos[0])
        elif pos[1] >= self.rect.h:
            return self.find_index(len(self.lines) - 1, self.lines[-1].find_index(pos[0]))
        else:
            for line_index, line in enumerate(self.lines):
                if pos[1] < line.rect.bottom:
                    return self.find_index(line_index, line.find_index(pos[0]))
        assert self.lines[-1].rect.bottom == self.rect.h, f"{self.lines[-1].rect.bottom} {self.rect.h}"
        raise Exception

    def find_index(self, line_index, char_index):
        """
        This method return the total index from a line index and a character index

        Example:
            text = "Hello\n"
                   "world"
            text.find_index(1, 2) -> index between 'o' and 'r'
                                  -> 8

        WARNING : this method result don't always match with text.index('r'), when
                  the text is cut inside a word or after a '-', we need two different
                  indexes for the end of the line and the start of the next line
        """

        text_index = 0
        for i, line in enumerate(self.lines):
            if i == line_index:
                break
            text_index += len(line.text_with_end)
        return text_index + char_index

    def _find_indexes(self, pos):
        """
        Renvoie l'index correspondant a l'espace entre deux caracteres le plus proche

        Exemple :
            pos = (40, 23)                             (40eme pixel a partir de self.left, 23eme pixel sous self.top)
            find_index(pos) -> self.find_index(2, 13)  (pos est sur le 14eme caractere de la 3eme ligne)
        """

        if pos[1] < 0:
            return 0, 0
        elif pos[1] >= self.rect.h:
            return len(self.lines) - 1, len(self.lines[-1].text)
        else:
            for line_index, line in enumerate(self.lines):
                if line.rect.bottom > pos[1]:
                    return line_index, line.find_index(pos[0])

        raise Exception

    def find_indexes(self, text_index):
        """
        Renvoie l'inverse de self.find_index(line_index, char_index)

        Example:
            text = "Hello\n"
                   "world"
            text.find_indexes(8) -> (1, 2) (index is between 'wo' & 'rld')
        """

        if text_index < 0:
            return 0, 0

        for line_index, line in enumerate(self.lines):
            if text_index <= len(line.text):
                return line_index, text_index
            text_index -= len(line.text_with_end)

        # The given text_index is too high
        return len(self.lines), len(self.lines[-1].text)

    def _find_mouse_index(self):
        """
        Return the closest index from mouse.x
        """
        return self._find_index(pos=self.abs_rect.referencing(mouse.pos))

    def get_text(self):
        return ''.join(line.text_with_end for line in self.lines)[:-1]  # Discard last \n

    text = property(get_text)

    def pack(self, *args, **kwargs):

        raise PermissionError("Should not use this method on a Text")

    def set_text(self, text):

        if self.has_locked("text"):
            return

        with paint_lock:

            for child in tuple(self.lines):
                assert child in self.children
                assert self == child.parent

                child.kill()

            line_class = _SelectableLine if self.is_selectable else _Line
            line_index = 0
            for paragraph_text in text.split("\n"):

                create_line = True
                while create_line:

                    line_text, other_lines_text = self._cut_text(paragraph_text)

                    if other_lines_text == '':
                        create_line = False
                        end = '\n'
                    else:
                        end = ''
                        paragraph_text = other_lines_text

                    line_class(
                        parent=self,
                        text=line_text,
                        line_index=line_index,  # TODO : usefull ?
                        end=end,
                    )
                    line_index += 1

            self._name = self.lines[0].text

            # if not self.height_is_adaptable and self.lines[-1].rect.bottom > self.content_rect.bottom:
            #     # NOTE : loops because self.font.config() calls self.set_text()
            #     # while self.lines[-1].rect.bottom > self.content_rect.bottom:
            #     if self.font.height == 2:
            #         raise ValueError(
            #             f"This text is too long for the text area : {text} (area={self.content_rect}), "
            #             f"{self.align_mode}, {self.rect.width}")
            #     self.font.config(height=self.font.height - 1)  # changing the font automatically updates the text

    # Selectable methods
    def check_select(self, selection_rect):

        if not self.is_selectable:
            return

        for line in self.lines:
            line.check_select(selection_rect)
        self._is_selected = True in tuple((line.is_selected for line in self.lines))

    def get_selected_data(self):

        if self.is_selected:
            return ''.join(line.get_selected_data() for line in self.lines)

    def handle_selector_link(self):

        if not self.is_selectable:
            return

        if not self.collidemouse():
            return

        if mouse.has_triple_clicked:
            for line in self.lines:
                if line.abs_rect.top <= mouse.y < line.abs_rect.bottom:
                    with paint_lock:
                        self.selector.close_selection()
                        self.selector.start_selection((line.abs_rect.left, line.abs_rect.top))
                        self.selector.end_selection((line.abs_rect.right, line.abs_rect.top), visible=False)
                        return
        elif mouse.has_double_clicked:
            for line in self.lines:
                if line.abs_rect.top <= mouse.y < line.abs_rect.bottom:
                    line.select_word(line.find_mouse_index())

    def handle_unselect(self):

        if not self.is_selectable:
            return

        for line in self.lines:
            if line.is_selected:
                line.handle_unselect()


class DynamicText(Text, Runable):

    def __init__(self, parent, get_text, **kwargs):
        assert callable(get_text), get_text

        Text.__init__(self, parent=parent, text=str(get_text()), **kwargs)
        Runable.__init__(self, parent)

        self._get_new_text = get_text

        self.set_running(True)

    def run(self):
        new_text = str(self._get_new_text())
        if new_text != self.text:
            self.set_text(new_text)




import pygame
from baopig.communicative import Communicative
from baopig.time.timer import RepeatingTimer
from .logging import LOGGER


class _Keyboard(Communicative):
    _signals = "KEYDOWN", "KEYUP"

    def __init__(self):

        Communicative.__init__(self)

        # 2 signals :
        #     - KEYDOWN
        #     - KEYUP
        for signal in self._signals:
            self.create_signal(signal)

        self._keys = {}
        self._application = None
        self.last_event = None

        class Mod:
            def __init__(self):
                self.l_alt = False
                self.r_alt = False
                self.alt = self.l_alt or self.r_alt
                self.l_cmd = False
                self.r_cmd = False
                self.cmd = self.l_cmd or self.r_cmd
                self.l_ctrl = False
                self.r_ctrl = False
                self.ctrl = self.l_alt or self.r_ctrl
                self.l_maj = False
                self.r_maj = False
                self.maj = self.l_maj or self.r_maj
            # def __str__(self):
            #     return Object.__str__(self)
        self._mod = Mod()

        # repeat
        self._pressedkeys_timers = {}  # the time when the key have been pressed
        self._is_repeating = False
        self._repeat_first_delay = None
        self._repeat_delay = None

        self.set_repeat(.5, .05)

    application = property(lambda self: self._application)
    is_repeating = property(lambda self: self._is_repeating)
    mod = property(lambda self: self._mod)

    def is_pressed(self, key):
        """Return True if the key with identifier 'key' (an integer) is pressed"""

        # You can write bp.keyboard.is_pressed("z")
        if isinstance(key, str):
            key = getattr(self, key)

        try:
            return bool(self._keys[key])
        except KeyError:
            # Here, the key has never been pressed
            return 0

    def receive(self, event):
        """Receive pygame events from the application"""

        self.last_event = event

        # ACTUALIZING KEYBOARD STATES
        if event.type == pygame.KEYDOWN:
            self._keys[event.key] = 1
            if event.key not in self._pressedkeys_timers:
                self._pressedkeys_timers[event.key] = None
            if self._is_repeating and self._pressedkeys_timers[event.key] is None:
                repeat = RepeatingTimer((self._repeat_first_delay, self._repeat_delay),
                                        pygame.event.post, event)
                repeat.start()
                self._pressedkeys_timers[event.key] = repeat
        elif event.type == pygame.KEYUP:
            if self._keys[event.key] == 0:
                return  # The KEYDOWN have been skipped, so we skip the KEYUP
            self._keys[event.key] = 0
            if self._is_repeating:
                assert self._pressedkeys_timers[event.key] is not None
                self._pressedkeys_timers[event.key].cancel()
                self._pressedkeys_timers[event.key] = None
        else:
            LOGGER.warning(f"Unexpected event : {event}")
            return

        if event.key == pygame.K_RALT:
            self.mod.r_alt = event.type == pygame.KEYDOWN
            self.mod.alt = self.mod.l_alt or self.mod.r_alt
        elif event.key == pygame.K_LALT:
            self.mod.l_alt = event.type == pygame.KEYDOWN
            self.mod.alt = self.mod.l_alt or self.mod.r_alt
        elif event.key == pygame.K_RMETA:
            self.mod.r_cmd = event.type == pygame.KEYDOWN
            self.mod.cmd = self.mod.l_cmd or self.mod.r_cmd
        elif event.key == pygame.K_LMETA:
            self.mod.l_cmd = event.type == pygame.KEYDOWN
            self.mod.cmd = self.mod.l_cmd or self.mod.r_cmd
        elif event.key == pygame.K_RCTRL:
            self.mod.r_ctrl = event.type == pygame.KEYDOWN
            self.mod.ctrl = self.mod.l_ctrl or self.mod.r_ctrl
        elif event.key == pygame.K_LCTRL:
            self.mod.l_ctrl = event.type == pygame.KEYDOWN
            self.mod.ctrl = self.mod.l_ctrl or self.mod.r_ctrl
        elif event.key == pygame.K_RSHIFT:
            self.mod.r_maj = event.type == pygame.KEYDOWN
            self.mod.maj = self.mod.l_maj or self.mod.r_maj
        elif event.key == pygame.K_LSHIFT:
            self.mod.l_maj = event.type == pygame.KEYDOWN
            self.mod.maj = self.mod.l_maj or self.mod.r_maj

        # EVENT TRANSMISSION
        for signal_id in self._signals:
            if event.type == getattr(pygame, signal_id):
                getattr(self.signal, signal_id).emit(event)
                break

    def set_repeat(self, first_delay, delay):
        """Control how held keys are repeated, with delays in seconds"""
        # This solves a bug in pygame, who can't repeat two keys

        assert first_delay >= 0
        assert delay > 0

        if 1 in self._keys:
            raise PermissionError("You must set the keys repeat before launch the application")

        self._is_repeating = True
        self._repeat_first_delay = first_delay
        self._repeat_delay = delay


keyboard = _Keyboard()



from baopig import *


class RainbowRect(Rectangle):
    def __init__(self, parent, **kwargs):
        self.paint_calls = 0
        Rectangle.__init__(self, parent, color=(255, 0, 0), **kwargs)

    def paint(self):
        self.paint_calls = (self.paint_calls + 3) % 360
        self.color.set_hue(self.paint_calls)
        super().paint()


class UT_Paintable_Zone(Zone):
    def __init__(self, *args, **kwargs):
        Zone.__init__(self, *args, **kwargs)

        Layer(self, name="zones_layer", spacing=10)
        z1 = Zone(self, background_color=(150, 150, 150), padding=10, spacing=10)
        self.default_layer.pack()

        rainbow = RainbowRect(z1)
        buttons_zone = Zone(z1, spacing=10)
        Button(buttons_zone, text="dirty = 0", command=PrefilledFunction(rainbow.set_dirty, 0))
        Button(buttons_zone, text="dirty = 1", command=PrefilledFunction(rainbow.set_dirty, 1))
        Button(buttons_zone, text="dirty = 2", command=PrefilledFunction(rainbow.set_dirty, 2))
        Button(buttons_zone, text="paint request", command=rainbow.send_paint_request)
        buttons_zone.pack(axis="horizontal", adapt=True)
        rainbow.resize(*buttons_zone.rect.size)
        z1.pack(adapt=True)

    def load_sections(self):
        self.parent.add_section(
            title="Paintable.set_dirty()",
            tests=[
                "if dirty is 0, paint() is not requested",
                "if dirty is 1, paint() is called at next frame rendering, then dirty will be set to 0 again",
                "if dirty is 2, paint() is called at each frame rendering",
            ]
        )
        self.parent.add_section(
            title="Paintable.send_paint_request()",
            tests=[
                "if dirty was 0, paint() is called at next frame rendering, then dirty will be set to 0 again",
            ]
        )


# For the PresentationScene import
ut_zone_class = UT_Paintable_Zone

if __name__ == "__main__":
    from baopig.prefabs.testerscene import TesterScene
    app = Application()
    TesterScene(app, ut_zone_class)
    app.launch()




import threading
from baopig.io import LOGGER


class ExtraThread(threading.Thread):

    def __init__(self, app):

        threading.Thread.__init__(self, name=self.__class__.__name__, daemon=True)

        self._application = app
        self._stop_event = threading.Event()
        self._stop_event.set()

    def __str__(self):

        return self.name

    app = property(lambda self: self._application)
    is_running = property(lambda self: not self._stop_event.is_set())
    is_stopped = property(lambda self: self._stop_event.is_set())

    def init(self):
        """
        Called when the thread restart
        """

    def run(self):

        while True:

            assert self.is_alive()
            if self.is_running:

                LOGGER.debug("Start to run thread : {}".format(self))
                self.init()

                while self.is_running:

                    try:
                        self.update()
                    except Exception as e:
                        LOGGER.exception("Exception on {} :  {}".format(self, e))

                LOGGER.debug("Stop to run thread : {}".format(self))

    def start(self):

        threading.Thread.start(self)
        self._stop_event.clear()

    def stop(self):

        self._stop_event.set()
        self.update = lambda: None

    def update(self):
        """
        Called as much as possible while the thread is running
        """



import pygame
from baopig.lib import Rectangle, Color, paint_lock
from baopig.widgets.slider import Slider, SliderBar, SliderBloc
from baopig.widgets.text import Text
from baopig.widgets.numentry import NumEntry
from baopig.widgets.dialog import Dialog, DialogFrame, DialogAnswerButton


class ColorSliderBloc(SliderBloc):

    def paint(self):
        self.surface.fill(self.border_color)
        pygame.draw.rect(
            self.surface, self.slider.color,
            (self.border_width, self.border_width,
             self.rect.w - self.border_width * 2, self.rect.h - self.border_width * 2)
        )


class ColorSliderBar(SliderBar):

    slider = property(lambda self: self._parent)

    def paint(self):

        color = self.slider.color.copy()
        if self.border_color is not None:
            pygame.draw.rect(self.surface, self.border_color, (0, 0) + self.rect.size, self.border_width * 2 - 1)
        bar_width = self.rect.w - self.border_width * 2
        for i in range(bar_width):
            if self.parent.attr in ("s", "v", "l"):
                color = self.slider.parent.color.copy()
            setattr(color, self.slider.attr, int(i / bar_width * (self.parent.maxval + 1)))
            pygame.draw.line(
                self.surface, color,
                (i + self.border_width, self.border_width),
                (i + self.border_width, self.rect.h - self.border_width * 2)
            )


class ColorSlider(Slider):
    STYLE = Slider.STYLE.substyle()
    STYLE.modify(
        length=260,
        loc="midleft",
        bloc_class=ColorSliderBloc,
        bar_class=ColorSliderBar,
    )

    def __init__(self, parent, attr, y, maxval=255, **kwargs):

        self.attr = attr
        self.color = parent.color.copy()
        self.color_before_link = None
        Slider.__init__(
            self, parent,
            minval=0, maxval=maxval, defaultval=getattr(parent.color, attr), step=1,
            pos=(parent.slider_x, y), **kwargs
        )
        parent.sliders.append(self)

        self.signal.NEW_VAL.connect(self.handle_new_val, owner=self)

    def handle_link(self):

        self.color_before_link = self.color.copy()
        super().handle_link()

    def handle_new_val(self, val):
        if self.is_linked:
            self.color = self.color_before_link.copy()
        setattr(self.color, self.attr, val)
        self.parent.update(self.color)

    def handle_unlink(self):

        self.update()

    def update(self):

        if self.is_linked:
            return self.bloc.send_paint_request()
        self.color = self.parent.color.copy()
        self._update_val(getattr(self.parent.color, self.attr))
        with paint_lock:
            self.bloc.send_paint_request()
            self.bar.send_paint_request()


class ColorEntry(NumEntry):
    STYLE = NumEntry.STYLE.substyle()
    STYLE.modify(
        loc="midleft",
        padding=2,
    )

    def __init__(self, parent, attr, maxval, y):
        NumEntry.__init__(
            self, parent,
            minval=0, maxval=maxval, accept_floats=False, default=getattr(parent.color, attr),
            pos=(parent.entry_x, y),
        )
        self.attr = attr
        self.parent.entries.append(self)

    def handle_validate(self):
        super().handle_validate()
        setattr(self.parent.color, self.attr, int(self.text))
        self.parent.update()

    def update(self):
        val = str(round(getattr(self.parent.color, self.attr)))
        self.set_text(val)


class ColorAnswerButton(DialogAnswerButton):
    def handle_validate(self):
        self.dialog._answer(self.dialog.frame.color)


class ColorDialogFrame(DialogFrame):
    STYLE = DialogFrame.STYLE.substyle()
    STYLE.modify(
        width=600,
        height=360,
    )

    def __init__(self, dialog, **kwargs):

        DialogFrame.__init__(self, dialog, **kwargs)
        self.color = Color(0, 0, 0)

        self.is_updating = False

        self.set_style_for(ColorSlider, length=256 + 2, wideness=13)

        self.label_x = 10
        self.entry_x = self.label_x + 5 + 90
        self.slider_x = self.entry_x + 5 + 45

        self.red_y = 100
        self.green_y = self.red_y + 20 + 5
        self.blue_y = self.green_y + 20 + 5
        self.rgb_y = [self.red_y, self.green_y, self.blue_y]
        self.hue_y = self.blue_y + 20 + 5
        self.saturation_y = self.hue_y + 20 + 5
        self.value_y = self.saturation_y + 20 + 5
        self.lightness_y = self.value_y + 20 + 5

        self.sliders = []
        self.entries = []

        pos = (self.slider_x + 300, self.red_y - 20)
        self.color_rect = Rectangle(
            self, color=self.color,
            size=(100, self.lightness_y + 20 - pos[1]),
            pos=pos
        )
        self.color_text = Text(
            self, str(self.color),
            midbottom=(0, -10),
            ref=self.color_rect, refloc="midtop"
        )

        if True:
            # RED
            self.red_label = Text(self, "Red :", midleft=(self.label_x, self.red_y))
            self.red_entry = ColorEntry(self, "r", 255, self.red_y)
            self.red_slider = ColorSlider(self, "r", self.red_y)

            # GREEN
            self.green_label = Text(self, "Green :", midleft=(self.label_x, self.green_y))
            self.green_entry = ColorEntry(self, "g", 255, self.green_y)
            self.green_slider = ColorSlider(self, "g", self.green_y)

            # BLUE
            self.blue_label = Text(self, "Blue :", midleft=(self.label_x, self.blue_y))
            self.blue_entry = ColorEntry(self, "b", 255, self.blue_y)
            self.blue_slider = ColorSlider(self, "b", self.blue_y)

        if True:
            # HUE
            self.hue_label = Text(self, "Hue :", midleft=(self.label_x, self.hue_y))
            self.hue_entry = ColorEntry(self, "h", 359, self.hue_y)
            self.hue_slider = ColorSlider(self, "h", self.hue_y, maxval=359)

            # SATURATION
            self.saturation_label = Text(self, "Saturation :", midleft=(self.label_x, self.saturation_y))
            self.saturation_entry = ColorEntry(self, "s", 100, self.saturation_y)
            self.saturation_slider = ColorSlider(self, "s", self.saturation_y, maxval=100)

            # SATURATION
            self.value_label = Text(self, "Value :", midleft=(self.label_x, self.value_y))
            self.value_entry = ColorEntry(self, "v", 100, self.value_y)
            self.value_slider = ColorSlider(self, "v", self.value_y, maxval=100)

            # SATURATION
            self.lightness_label = Text(self, "Lightness :", midleft=(self.label_x, self.lightness_y))
            self.lightness_entry = ColorEntry(self, "l", 100, self.lightness_y)
            self.lightness_slider = ColorSlider(self, "l", self.lightness_y, maxval=100)

    def update(self, color=None):

        if self.is_updating:
            return  # avoid update loops
        self.is_updating = True
        if color:
            self.color = color
        for entry in self.entries:
            entry.update()
        for slider in self.sliders:
            slider.update()
        self.color_rect.set_color(self.color)
        self.color_text.set_text(str(self.color))
        self.is_updating = False


class ColorChooserDialog(Dialog):
    STYLE = Dialog.STYLE.substyle()
    STYLE.modify(
        title="Color Chooser",
        choices=("Select",),
        dialogframe_class=ColorDialogFrame,
        answerbutton_class=ColorAnswerButton,
    )



import functools
import inspect
from collections import deque
from weakref import ref as wkref, WeakSet


def ref(obj, callback=None):
    if hasattr(obj, "get_weakref"):
        return obj.get_weakref(callback)
    return wkref(obj, callback)


class PrefilledFunction:
    def __init__(self, function, *args, **kwargs):
        assert callable(function), "function must be callable"

        self._function = function
        self._args = args
        self._kwargs = kwargs

    def __call__(self, *args, **kwargs):
        return self._function(*self._args, *args, **self._kwargs, **kwargs)


class PackedFunctions:
    """
    This object allow tou to execute multiple function from one call

    WARNING : if you enter parameters, it will be distributed to all of the packed functions
    This means you cannot pack two functions if they each require differents parameters
    You cannot remove a function from a PackedFunctions
    """

    def __init__(self, *functions):

        self._functions = list()
        for func in functions:
            self.add(func)

    def __call__(self, *args, **kwargs):

        for func in self._functions:
            func(*args, **kwargs)

    def __str__(self):

        return "PackedFunctions{}".format(tuple(self._functions))

    def add(self, func, owner=None):
        """
        Owner parameter is usefull for remove
        It allow you to remove an element who own a method in a PackedFunctions
        because if you don't, the owner will not be deleted
        """

        assert callable(func)
        if owner is not None:
            setattr(func, "_owner", owner)
            owner.listof_packedfunctions.add(self)
        self._functions.append(func)

    def clear(self):

        self._functions.clear()

    def remove(self, func):

        self._functions.remove(func)

    def remove_funcs_of(self, owner):

        # funcs_to_remove = []
        for func in tuple(self._functions):
            if hasattr(func, "_owner"):
                if func._owner == owner:
                    # funcs_to_remove.append(func)
                    self._functions.remove(func)
        # for func in funcs_to_remove:


class Object:
    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            self.__setattr__(key, value)

    def __str__(self) -> str:
        return "<{}({})>".format(self.__class__.__name__, str(self.__dict__)[1:-1])


class TypedDeque(deque):
    """
    A TypedDeque is a deque who can only contain items of type ItemsClass
    deque : a list-like sequence optimized for data accesses near its endpoints
    """

    def __init__(self, ItemsClass, seq=(), maxlen=None):
        """
        :param ItemsClass: the type for all deque items
        :param seq: an items sequence
        :param maxlen: the max deque length
        """

        assert inspect.isclass(ItemsClass), "ItemsClass must be a class"
        assert hasattr(seq, "__iter__"), "seq must be an iterable"
        assert maxlen is None or isinstance(maxlen, int), "maxlen must be an integer"

        self.ItemsClass = ItemsClass
        self.msg_item_type_error = "Only {} objects are accepted in this list".format(self.ItemsClass.__name__)

        for item in seq:
            self._check(item)
        deque.__init__(self, seq, maxlen)

    def __setitem__(self, index, p_object):
        self._check(p_object)

        deque.__setitem__(self, index, p_object)

    def __repr__(self):
        return "<TypedDeque({}):{}>".format(self.ItemsClass.__name__,
                                            "[{}]".format(", ".join(list((item.__str__() for item in self)))))

    def __str__(self):
        return "[{}]".format(", ".join(list((item.__str__() for item in self)))) + (
            ", maxlen={}".format(self.maxlen) if self.maxlen else "")

    def _check(self, p_object):
        if not self.accept(p_object):
            raise PermissionError(self.msg_item_type_error)

    def accept(self, p_object):
        return isinstance(p_object, self.ItemsClass)

    def append(self, p_object):
        self._check(p_object)

        deque.append(self, p_object)

    def appendleft(self, p_object):
        self._check(p_object)

        deque.appendleft(self, p_object)

    def extend(self, iterable):
        for p_object in iterable:
            self.append(p_object)

    def extendleft(self, iterable):
        for p_object in iterable:
            self.appendleft(p_object)

    def insert(self, index, p_object):
        self._check(p_object)

        deque.insert(self, index, p_object)


class TypedList(list):

    def __init__(self, *ItemsClass, seq=()):
        """
        Create a list who can only contain items of type ItemsClass
        :param ItemsClass: the type for all list items
        :type ItemsClass: class
        :param seq: an items sequence
        """
        list.__init__(self, seq)
        self.set_ItemsClass(*ItemsClass)

    def __setitem__(self, index, p_object):
        self._check(p_object)

        list.__setitem__(self, index, p_object)

    def __repr__(self):
        return "<{}(ItemsClass:{}, {}>".format(self.__class__.__name__, self.ItemsClass_name,
                                               "[{}]".format(", ".join(list((item.__str__() for item in self)))))

    def __str__(self):
        return "[{}]".format(", ".join(list((item.__str__() for item in self))))

    def _check(self, item):
        if not self.accept(item):
            raise PermissionError(self.msg_item_type_error.format(item.__class__.__name__))

    def accept(self, item):
        if inspect.isclass(item):
            return issubclass(item, self.ItemsClass)
        return isinstance(item, self.ItemsClass)

    def append(self, p_object):
        self._check(p_object)

        list.append(self, p_object)

    def extend(self, iterable):
        for p_object in iterable:
            self.append(p_object)

    def insert(self, index, p_object):
        self._check(p_object)

        list.insert(self, index, p_object)

    def set_ItemsClass(self, *ItemsClass):

        # try:
        for Class in ItemsClass:
            assert inspect.isclass(Class), "ItemsClass must be a class or a list of class"
        name = "(%s)" % ", ".join(Class.__name__ for Class in ItemsClass)
        # except TypeError:
        #     assert inspect.isclass(ItemsClass), "ItemsClass must be a class or a list of class"
        #     name = ItemsClass.__name__
        self.ItemsClass = ItemsClass
        self.ItemsClass_name = name
        self.msg_item_type_error = "Only {} objects are accepted in this list".format(self.ItemsClass_name) + \
                                   " (wrong object class:{})"
        for item in self:
            self._check(item)


class TypedSet(set):
    """
    A TypedSet is a unordered collection of unique elements
    who can only contain items of type ItemsClass

    seq is the optionnal initial sequence
    """

    def __init__(self, ItemsClass, seq=()):

        assert inspect.isclass(ItemsClass), "ItemsClass must be a class"
        assert hasattr(seq, "__iter__"), "seq must be an iterable"
        self.ItemsClass = ItemsClass
        self.msg_item_type_error = "Only {} objects are accepted in this list".format(self.ItemsClass.__name__)
        for item in seq:
            self._check(item)
        set.__init__(self, seq)

    def __repr__(self):
        return super().__repr__()
        # return "<TypedSet({}):{}>".format(self.ItemsClass.__name__, set.__str__(self))

    def __str__(self):
        return super().__str__()
        # return "{{}}".format(", ".join(list((item.__str__() for item in self))))

    def _check(self, item):
        if not self.accept(item):
            raise PermissionError(self.msg_item_type_error)

    def accept(self, item):
        return isinstance(item, self.ItemsClass)

    def add(self, item):
        """
        Add an item to a set.
        This has no effect if the item is already present.

        :param item: an item of type self.ItemsClass
        :return: None
        """
        self._check(item)

        set.add(self, item)

    def update(self, *args):
        """
        Add all items from args into a set
        if the items already is in the set, it isn't added

        S.update({1, 2, 3})
        S.update({1}, {2}, {3})

        :param args: a sequence of items sequence
        :return: None
        """
        for seq in args:
            for item in seq:
                self.add(item)


class History(deque):
    """
    An History is a TypedDeque whith a fixed size
    You can only :
        - append a new element in the history
        - read the history
    When you append a new element to the history, if it is full,
    the oldest element is removed
    The oldest element is positionned to the left

    Exemple :

            print(history) -> [2, 3, 4, 5, 6, 7, 8, 9]
                               ^                    ^
                        oldest element       newest element
    """

    def __init__(self, maxlen, seq=()):
        deque.__init__(self, seq, maxlen)

    def __delitem__(self, *args, **kwargs):
        raise PermissionError("Cannot use __delitem__ on an History")

    def __iadd__(self, *args, **kwargs):
        raise PermissionError("Cannot use __iadd__ on an History")

    def __imul__(self, *args, **kwargs):
        raise PermissionError("Cannot use __imul__ on an History")

    def __setitem__(self, index, p_object):
        raise PermissionError("Cannot use __setitem__ on an History")

    def appendleft(self, p_object):
        raise PermissionError("Cannot use appendleft on an History")

    def extendleft(self, iterable):
        raise PermissionError("Cannot use extendleft on an History")

    def insert(self, index, p_object):
        raise PermissionError("Cannot use insert on an History")

    def pop(self, *args, **kwargs):
        raise PermissionError("Cannot use pop on an History")

    def popleft(self, *args, **kwargs):
        raise PermissionError("Cannot use popleft on an History")

    def reverse(self):
        raise PermissionError("Cannot use reverse on an History")

    def rotate(self, *args, **kwargs):
        raise PermissionError("Cannot use rotate on an History")


class WeakList(list):
    """
    Create a TypedList who only store weak references to objects
    code from : https://stackoverflow.com/questions/677978/weakref-list-in-python
    """

    def __init__(self, seq=()):
        list.__init__(self)
        self._refs = []
        self._dirty = False
        for x in seq:
            self.append(x)

    def _mark_dirty(self, _):
        self._dirty = True

    def flush(self):
        self._refs = [x for x in self._refs if x() is not None]
        self._dirty = False

    def __eq__(self, other):

        return self is other

    def __getitem__(self, idx):
        if self._dirty:
            self.flush()
        if type(idx) == slice:
            return list(ref() for ref in self._refs[idx])
        return self._refs[idx]()

    def __iter__(self):
        if self._dirty:
            self.flush()
        for ref in self._refs:
            yield ref()

    def __repr__(self):
        return "WeakList(%r)" % list(self)

    def __str__(self):
        return "[{}]".format(", ".join(list((item.__str__() for item in list(self)))))

    def __len__(self):
        if self._dirty:
            self.flush()
        return len(self._refs)

    def __setitem__(self, idx, obj):
        if isinstance(idx, slice):
            self._refs[idx] = [ref(obj, self._mark_dirty) for _ in obj]
        else:
            self._refs[idx] = ref(obj, self._mark_dirty)

    def __delitem__(self, idx):
        del self._refs[idx]

    def append(self, obj):
        self._refs.append(ref(obj, self._mark_dirty))

    def count(self, obj):
        return list(self).count(obj)

    def extend(self, items):
        for x in items:
            self.append(x)

    def index(self, obj, **kwargs):
        return list(self).index(obj, **kwargs)

    def insert(self, idx, obj):
        self._refs.insert(idx, ref(obj, self._mark_dirty))

    def pop(self, *args, **kwargs):
        if self._dirty:
            self.flush()
        idx = args[0]
        obj = self._refs[idx]()
        del self._refs[idx]
        self.flush()
        return obj

    def remove(self, obj):
        if self._dirty:
            self.flush()  # Ensure all valid.
        for i, x in enumerate(self):
            if x == obj:
                del self[i]
        self.flush()

    def reverse(self):
        self._refs.reverse()

    def sort(self, key=None, reverse=False):
        if self._dirty:
            self.flush()
        if key is not None:
            key = lambda x, key=key: key(x())
        else:
            key = lambda x: x()
        self._refs.sort(key=key, reverse=reverse)

    def __add__(self, other):
        wl = WeakList(self)
        wl.extend(other)
        return wl

    def __iadd__(self, other):
        self.extend(other)
        return self

    def __contains__(self, obj):
        return obj in list(self)

    def __mul__(self, n):
        return WeakList(list(self) * n)

    def __imul__(self, n):
        self._refs *= n
        return self

    def __reversed__(self):
        for ref in self._refs.__reversed__():
            yield ref()


class WeakTypedList(TypedList, WeakList):

    def __init__(self, *ItemsClass, seq=()):
        WeakList.__init__(self, seq=seq)
        TypedList.__init__(self, *ItemsClass)

    def __eq__(self, other):
        return self is other

    def __setitem__(self, index, p_object):
        self._check(p_object)

        WeakList.__setitem__(self, index, p_object)

    def __str__(self):
        return WeakList.__str__(self)

    def append(self, p_object):
        self._check(p_object)

        WeakList.append(self, p_object)

    def extend(self, iterable):
        for p_object in iterable:
            self.append(p_object)

    def insert(self, index, p_object):
        self._check(p_object)

        WeakList.insert(self, index, p_object)


class WeakTypedSet(WeakSet, TypedSet):
    """
    A TypedSet is a unordered collection of unique elements
    who can only contain items of type ItemsClass

    seq is the optionnal initial sequence

    WARNING : elements of a WeakTypedSet object are stored in WeakTypedSet.data
    """

    def __init__(self, ItemsClass, seq=()):

        WeakSet.__init__(self, data=seq)
        TypedSet.__init__(self, ItemsClass=ItemsClass)

    def __eq__(self, other):

        return self is other

    def __repr__(self):
        return super().__repr__()
        # return "<WeakTypedSet({}):{}>".format(self.ItemsClass.__name__, set.__str__(self))

    def __str__(self):
        return super().__str__()
        # return "{{}}".format(", ".join(list((item.__str__() for item in self))))

    def add(self, item):
        """
        Add an item to a set.
        This has no effect if the item is already present.

        :param item: an item of type self.ItemsClass
        :return: None
        """
        self._check(item)
        WeakSet.add(self, item)

    def update(self, *args):
        """
        Add all items from args into a set
        if the items already is in the set, it isn't added

        S.update({1, 2, 3})
        S.update({1}, {2}, {3})

        :param args: a sequence of items sequence
        :return: None
        """
        for seq in args:
            for item in seq:
                self.add(item)


def get_name(obj):
    for k, v in globals().items():
        if k == "obj":
            continue
        if v == obj:
            return k
    return None


def debug(func):
    functools.wraps(func)

    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            raise e

    return wrapper




import threading
import pygame


# TODO : deproteger les attributs handle_something


class MetaPaintLocker(type):
    def __call__(cls, *args, **kwargs):
        with paint_lock:
            widget = super().__call__(*args, **kwargs)
            if widget.is_awake:
                widget.parent._add_child(widget)
            return widget


class Color(pygame.Color):
    """
    Very close to pygame.Color :

    Color(tuple) -> Color
    Color(name) -> Color
    Color(r, g, b, a) -> Color
    Color(rgbvalue) -> Color
    pygame object for color representations
    """

    def __init__(self, *args, transparency=None, **kwargs):

        if transparency is None:
            try:
                pygame.Color.__init__(self, *args, **kwargs)
            except ValueError:
                pygame.Color.__init__(self, *args[0])

        else:
            try:
                color = pygame.Color(*args, **kwargs)
            except ValueError:
                color = pygame.Color(*args[0])
            pygame.Color.__init__(self, color.r, color.g, color.b, transparency)

    def set_hsv(self, val):
        if val[1] > 100:
            val = val[0], 100, val[2]
        try:
            self.hsva = val + (100.,)
        except ValueError as e:
            raise ValueError(str(e) + f" : {val}")
    hsv = property(lambda self: self.hsva[:-1], set_hsv)

    def set_hsl(self, val):
        if val[1] > 100:
            val = val[0], 100, val[2]
        try:
            self.hsla = val + (100.,)
        except ValueError as e:
            raise ValueError(str(e) + f" : {val}")
    hsl = property(lambda self: self.hsla[:-1], set_hsl)

    def set_hue(self, h):
        self.hsv = (h,) + self.hsv[1:]
    h = property(lambda self: self.hsva[0], set_hue)  # TODO : replace h by hue

    def get_saturation(self):
        s = self.hsva[1]
        if s > 100:  # sometime 100.00000000000001
            s = 100
        return s

    def set_saturation(self, s):
        self.hsv = self.h, s, self.v

    s = property(get_saturation, set_saturation)

    def set_value(self, v):
        self.hsv = self.hsva[:2] + (v,)

    v = property(lambda self: self.hsva[2], set_value)

    def set_lightness(self, lightness):
        self.hsl = self.hsla[:2] + (lightness,)

    l = property(lambda self: self.hsla[2], set_lightness)

    def copy(self):
        return self.__class__(self)

    def has_transparency(self):

        return self.a < 255


class MarginType:
    # Once it is created, it can never change

    def __init__(self, margin):

        if margin is None:
            margin = 0, 0, 0, 0
        elif isinstance(margin, int):
            margin = tuple([margin] * 4)
        elif isinstance(margin, MarginType):
            margin = margin.left, margin.top, margin.right, margin.bottom
        else:
            ln = len(margin)
            if ln == 2:
                margin = tuple(margin) + tuple(margin)
            elif ln == 3:
                margin = tuple(margin) + tuple([margin[1]])
            else:
                assert ln == 4, f"Wrong value for margin type : {margin}"

        self._left = margin[0]
        self._top = margin[1]
        self._right = margin[2]
        self._bottom = margin[3]

    def __repr__(self):
        return f"MarginType(left={self.left}, top={self.top}, right={self.right}, bottom={self.bottom})"

    bottom = property(lambda self: self._bottom)
    is_null = property(lambda self: self.left == self.top == self.right == self.bottom == 0)
    left = property(lambda self: self._left)
    right = property(lambda self: self._right)
    top = property(lambda self: self._top)


class Handler_SceneOpen:
    """
    A Handler_SceneOpen is a widget whose 'handle_scene_open' function is called when its scene gets open
    """
    def handle_scene_open(self):
        """Stuff to do when the widget's scene gets open"""


class Handler_SceneClose:
    """
    A Handler_SceneClose is a widget whose 'handle_scene_close' function is called when its scene is closed
    """

    def handle_scene_close(self):
        """Stuff to do when the widget'scene is closed"""


paint_lock = threading.RLock()





from baopig import *


class UT_Scrollable_Zone(Zone):
    def __init__(self, *args, **kwargs):
        Zone.__init__(self, *args, padding=10, spacing=10, **kwargs)

        self.set_style_for(Button, height=20)

        TextEdit(self, text="Bonjour tout le monde.\n\n\n\n\n\n\n\n\n\n\nHello\n\n\n\nHello 2",
                 size=(300, 200), background_color="pink")
        Button(self)

        self.pack()


# For the PresentationScene import
ut_zone_class = UT_Scrollable_Zone

if __name__ == "__main__":
    from baopig.prefabs.testerscene import TesterScene
    app = Application()
    TesterScene(app, ut_zone_class)
    app.launch()



def is_one_of_types(o, *types):
    return isinstance(o, types)
    # return True in [isinstance(o, t) for t in types], f"{o} must be one of {types} types, got {type(o)} instead"


def is_iterable(o, k):
    if not hasattr(o, "__iter__"):
        return False
    return len(o) == k


def is_typed_iterable(o, t, k):
    if not hasattr(o, "__iter__"):
        return False
    if not len(o) == k:
        return False
    for elt in o:
        if not isinstance(elt, t):
            return False
    return True


def is_point(o):
    if not hasattr(o, "__iter__"):
        return False
    if not len(o) == 2:
        return False
    for coord in o:
        if not is_one_of_types(coord, int, float):
            return False
    return True


def is_size(o):
    if not hasattr(o, "__iter__"):
        return False
    if not len(o) == 2:
        return False
    for coord in o:
        if not is_one_of_types(coord, int, float):
            return False
        if not coord >= 0:
            return False
    return True


def is_color(o):
    if not hasattr(o, "__iter__"):
        return False
    if not len(o) in (3, 4):
        return False
    for elt in o:
        if not isinstance(elt, int):
            return False
        if not 0 <= elt <= 255:
            return False
    return True



from .scrollview import ScrollView
from .text import DynamicText, Text
from .textedit import TextEdit
from .lineedit import LineEdit
from .entry import Entry
from .numentry import NumEntry
from .button import Button, Button_Text

# TODO : rework ?
from .button import AButton_FocusSail as Button_FocusSail
from .button import AButton_HoverSail as Button_HoverSail
from .button import AButton_LinkSail as Button_LinkSail

from .checkbox import CheckBox
from .progressbar import ProgressBar
from .slider import Slider, SliderBloc, SliderBar
from .dialog import Dialog, DialogFrame, DialogButtonsZone, DialogAnswerButton
from .indicator import Indicator, DynamicIndicator
